import numpy
import tqdm
from scipy.spatial import cKDTree
from collections import defaultdict


def extract_faces(surface, mesh, crease_angle: float = 60, verbose: bool = False, combine_walls: bool = True):
    """
    This function extracts the boundary domains from a given
    surface mesh file in PolyData format. By default, a
    separation angle is assumed to divide the C0 continuous
    manifold along disjoint C1 discontinuities which fail to
    meet the angle threshold criteria.

    Parameters
    ----------
        surface : PolyData surface mesh
        mesh : UnstructuredGrid
        crease_angle : float
        verbose : bool (default: False)
    Returns
    -------
        faces : dict
        iscap : list
        wall_surfaces: list of PolyData
        cap_surfaces: list of PolyData
    """
    face_vertices = surface.faces.reshape(-1, 4)[:, 1:]
    unallocated_elements = set(range(face_vertices.shape[0]))
    vertex_map = build_vertex_map(face_vertices)
    edge_neighbors = build_edge_map(face_vertices, vertex_map)
    #face_neighbors = build_face_neighbors(surface)
    element_quality = surface.compute_cell_quality(quality_measure='scaled_jacobian')
    element_normals = surface.compute_normals(non_manifold_traversal=False, auto_orient_normals=True).cell_data["Normals"]
    point_normals = surface.compute_normals(non_manifold_traversal=False, auto_orient_normals=True).point_data["Normals"]
    element_normals = correct_normals(element_normals, face_vertices, surface)
    collapsed_elements = numpy.isclose(element_quality.cell_data["CellQuality"], 0.0, atol=1e-3)
    faces = partition(unallocated_elements, face_vertices, element_normals, crease_angle, vertex_map, edge_neighbors,
                      collapsed_elements)
    # Add a reconnecting step to anneal partitioned faces that are degenerate or collapsed into faces
    face_trees = []
    face_boundaries = []
    for face in faces:
        face_trees.append(cKDTree(surface.points[face_vertices[face, :].flatten(), :]))
        face_boundary = surface.extract_cells(face).extract_surface().extract_feature_edges(boundary_edges=True,
                                                                                             manifold_edges=False,
                                                                                             feature_edges=False,
                                                                                             non_manifold_edges=False)
        face_boundaries.append([cKDTree(body.points) for body in face_boundary.split_bodies()])
    new_faces = []
    new_idx = []
    for i, face in enumerate(faces):
        face_cells = surface.extract_cells(face).extract_surface()
        face_boundary = face_cells.extract_feature_edges(boundary_edges=True, manifold_edges=False,
                                                         feature_edges=False, non_manifold_edges=False)
        annealed = False
        if face_cells.n_points == face_boundary.n_points and len(face_boundary.split_bodies()) == 1:
            if verbose:
                print("Face {} is degenerate with no internal vertices.".format(i))
            for j, bounds in enumerate(face_boundaries):
                if i == j:
                    continue
                for k, bound in enumerate(bounds):
                    dists, _ = bound.query(face_boundary.points)
                    if numpy.all(numpy.isclose(dists, 0.0)):
                        if verbose:
                            print("Degenerate face {} annealed to face -> {}".format(i, j))
                        faces[j].extend(face)
                        annealed = True
                        break
                if annealed:
                    break
        #for j, tree in enumerate(face_trees):
        #    if i == j:
        #        continue
        #    dists, _ = tree.query(face_cells.points)
        #    if numpy.all(numpy.isclose(dists, 0.0)):
        #        if verbose:
        #            print("Degenerate face {} annealed to face -> {}".format(i, j))
        #        faces[j].extend(face)
        #        annealed = True
        #        break
        if not annealed:
            if verbose:
                print("Face {} is complete with internal vertices.".format(i))
            new_idx.append(i)
    for i in new_idx:
        new_faces.append(faces[i])
    faces = new_faces
    iscap = []
    wall_faces = []
    cap_faces = []
    wall_boundaries = []
    wall_boundary_trees = []
    cap_boundaries = []
    cap_boundary_trees = []
    for i in range(len(faces)):
        f = surface.extract_cells(faces[i]).extract_surface()
        tmp_bound_check = f.extract_feature_edges(boundary_edges=True, manifold_edges=False,
                                                 feature_edges=False, non_manifold_edges=False)
        tmp_bound_check = tmp_bound_check.split_bodies()
        #if numpy.sum(abs(numpy.max(f.cell_normals, axis=0) - numpy.min(f.cell_normals, axis=0))) < 0.1 or len(tmp_bound_check) == 1:
        if len(tmp_bound_check) == 1:
            iscap.append(1)
            cap_faces.append(faces[i])
            tmp_cap_boundaries = []
            tmp_cap_boundary_trees = []
            tmp_cap_boundaries.append(tmp_bound_check[0])
            tmp_cap_boundary_trees.append(cKDTree(tmp_bound_check[0].points))
        elif len(tmp_bound_check) > 1:
            tmp_wall_boundaries = []
            tmp_wall_boundary_trees = []
            boundaries = f.extract_feature_edges(boundary_edges=True, manifold_edges=False,
                                                 feature_edges=False, non_manifold_edges=False)
            boundaries = boundaries.split_bodies()
            for j in range(boundaries.n_blocks):
                tmp_wall_boundaries.append(boundaries[j])
                tmp_wall_boundary_trees.append(cKDTree(boundaries[j].points))
            wall_boundaries.append(tmp_wall_boundaries)
            wall_boundary_trees.append(tmp_wall_boundary_trees)
            iscap.append(0)
            wall_faces.append(faces[i])
        else:
            print("Error: Face {} has an unknown boundary edges??.".format(i))
    # Combine walls into single surfaces using matched boundary edges
    combined_walls = []
    new_wall_faces = []
    new_wall_idx = []
    for i in range(len(wall_faces)):
        if i not in combined_walls:
            for j in range(i + 1, len(wall_faces)):
                if j in combined_walls:
                    continue
                matched_boundary = False
                for k in range(len(wall_boundary_trees[i])):
                    for l in range(len(wall_boundary_trees[j])):
                        dists, _ = wall_boundary_trees[i][k].query(wall_boundary_trees[j][l].data)
                        complete_matches = []
                        if numpy.all(numpy.isclose(dists, 0.0)):
                            #wall_faces[i].extend(wall_faces[j])
                            matched_boundary = True
                            #combined_walls.append(j)
                            # Check to see if all boundaries are matched
                            for m in range(len(wall_boundary_trees[j])):
                                for ii in range(len(wall_boundary_trees)):
                                    if ii == j:
                                        continue
                                    for kk in range(len(wall_boundary_trees[ii])):
                                        dists, _ = wall_boundary_trees[j][m].query(wall_boundary_trees[ii][kk].data)
                                        if numpy.all(numpy.isclose(dists, 0.0)) or numpy.all(~numpy.isclose(dists, 0.0)):
                                            complete_matches.append(True)
                                        else:
                                            complete_matches.append(False)
                                            break
                                    if not all(complete_matches):
                                        break
                                for ii in range(len(cap_boundary_trees)):
                                    for kk in range(len(cap_boundary_trees[ii])):
                                        dists, _ = wall_boundary_trees[j][m].query(cap_boundary_trees[ii][kk].data)
                                        if numpy.all(numpy.isclose(dists, 0.0)) or numpy.all(~numpy.isclose(dists, 0.0)):
                                            complete_matches.append(True)
                                        else:
                                            print('Boundary Split on Cap {}'.format(ii))
                                            complete_matches.append(False)
                                            break
                                    if not all(complete_matches):
                                        break
                                if not all(complete_matches):
                                    break
                            if not all(complete_matches):
                                break
                        #if matched_boundary:
                        #    wall_faces[i].extend(wall_faces[j])
                        #    combined_walls.append(j)
                        #    break
                    if matched_boundary and combine_walls:
                        completely_matched_boundaries = all(complete_matches)
                        if completely_matched_boundaries:
                            wall_faces[i].extend(wall_faces[j])
                            combined_walls.append(j)
                            if verbose:
                                print("Wall {} combined into wall -> {}".format(j, i))
                            break
            new_wall_idx.append(i)
    new_faces = []
    new_iscap = []
    for i in new_wall_idx:
        new_faces.append(wall_faces[i])
        new_iscap.append(0)
    for i in range(len(cap_faces)):
        new_faces.append(cap_faces[i])
        new_iscap.append(1)
    faces = new_faces
    iscap = new_iscap
    walls = []
    caps = []
    for i, cap in enumerate(iscap):
        if not cap == 1:
            walls.append(faces[i])
        else:
            caps.append(faces[i])
    if not isinstance(mesh, type(None)):
        global_nodes = mesh.points
        global_node_tree = cKDTree(global_nodes)
        global_elements = mesh.cell_connectivity.reshape(-1, 4)
        global_elements = numpy.sort(global_elements, axis=1)
        tet_faces = []
        for i in tqdm.trange(global_elements.shape[0], desc="Building tetrahedral faces", leave=False):
            for j in range(4):
                idx = set(list(range(4))) - set([j])
                tet_faces.append(global_elements[i, list(idx)])
        tet_face_tree = cKDTree(tet_faces)
    #for i, cap in enumerate(iscap):
    #    if not cap == 1:
    #        walls.append(faces[i])
    #    else:
    #        caps.append(faces[i])
    # Map the surface ids to the mesh ids for wall
    wall_surfaces = []
    wall_boundaries = []
    for i in tqdm.trange(len(walls), desc="Mapping wall surfaces <-> mesh ids", leave=False):
        wall_cells = surface.extract_cells(walls[i])
        wall_surface = wall_cells.extract_surface()
        if not isinstance(mesh, type(None)):
            wall_surface.point_data["GlobalNodeID"] = numpy.zeros(wall_surface.n_points, dtype=int)
            wall_surface.cell_data['GlobalElementID'] = numpy.zeros(wall_surface.n_cells, dtype=int)
            wall_surface.cell_data['ModelFaceID'] = numpy.ones(wall_surface.n_cells, dtype=int)
            _, indices = global_node_tree.query(wall_surface.points)
            wall_surface.point_data["GlobalNodeID"] = indices.astype(int)
            # Assign Global Element IDs
            wall_faces = wall_surface.point_data["GlobalNodeID"][wall_surface.faces]
            wall_faces = wall_faces.reshape(-1, 4)[:, 1:]
            wall_faces = numpy.sort(wall_faces, axis=1)
            _, indices = tet_face_tree.query(wall_faces)
            wall_surface.cell_data["GlobalElementID"] = indices // 4
            wall_surface.cell_data["GlobalElementID"] = wall_surface.cell_data["GlobalElementID"].astype(int)
        boundaries = wall_surface.extract_feature_edges(boundary_edges=True, manifold_edges=False,
                                                             feature_edges=False, non_manifold_edges=False)
        boundaries = boundaries.split_bodies()
        tmp = []
        for j in range(boundaries.n_blocks):
            tmp.append(cKDTree(boundaries[j].points))
        wall_surfaces.append(wall_surface)
        wall_boundaries.append(tmp)
    # Map the surface ids to the mesh ids for the caps
    cap_surfaces = []
    cap_boundaries = []
    for i in tqdm.trange(len(caps), desc="Mapping cap surfaces <-> mesh ids", leave=False):
        face_cap = caps[i]
        cap_cells = surface.extract_cells(face_cap)
        cap_surface = cap_cells.extract_surface()
        if not isinstance(mesh, type(None)):
            cap_surface.point_data["GlobalNodeID"] = numpy.zeros(cap_surface.n_points, dtype=int)
            cap_surface.cell_data["GlobalElementID"] = numpy.zeros(cap_surface.n_cells, dtype=int)
            cap_surface.cell_data["ModelFaceID"] = numpy.ones(cap_surface.n_cells, dtype=int) * (i + 2)
            # Assign Global Node IDs
            _, indices = global_node_tree.query(cap_surface.points)
            cap_surface.point_data["GlobalNodeID"] = indices.astype(int)
            # Assign Global Element IDs
        boundaries = cap_surface.extract_feature_edges(boundary_edges=True, manifold_edges=False,
                                                           feature_edges=False, non_manifold_edges=False)
        boundaries = boundaries.split_bodies()
        tmp = []
        for j in range(boundaries.n_blocks):
            tmp.append(cKDTree(boundaries[j].points))
        cap_boundaries.append(tmp)
        cap_surfaces.append(cap_surface)
    shared_boundaries = {}
    for i in tqdm.trange(len(cap_boundaries), desc="Mapping shared boundaries", leave=False):
        shared_boundaries[i] = []
        for j in range(len(cap_boundaries[i])):
            for k in range(len(wall_boundaries)):
                for l in range(len(wall_boundaries[k])):
                    dists, _ = cap_boundaries[i][j].query(wall_boundaries[k][l].data)
                    if numpy.all(numpy.isclose(dists, 0.0)):
                        shared_boundaries[i].append(k)
                        break
    return faces, wall_surfaces, cap_surfaces, shared_boundaries


def build_vertex_map(face_vertices):
    n_points = int(numpy.max(face_vertices) + 1)
    vertex_map = [[] for i in range(n_points)]
    for i in tqdm.trange(face_vertices.shape[0], desc="Building vertex map", leave=False):
        for j in range(face_vertices.shape[1]):
            vertex_map[int(face_vertices[i, j])].append(i)
    for i in range(len(vertex_map)):
        vertex_map[i] = tuple(vertex_map[i])
    vertex_map = tuple(vertex_map)
    return vertex_map


def build_edge_map(face_vertices, vertex_map):
    neighbors = {}
    for i in tqdm.trange(face_vertices.shape[0], desc="Building edge map", leave=False):
        face = face_vertices[i, :]
        num_points = len(face)
        tmp = []
        for j in range(num_points):
            vrt_1 = set(vertex_map[face[j]])
            vrt_2 = set(vertex_map[face[(j + 1) % num_points]])
            vrt = vrt_1.intersection(vrt_2)
            vrt.remove(i)
            for k in vrt:
                tmp.append(k)
        neighbors[i] = tuple(tmp)
    return neighbors


def build_face_map(surface):
    face_map = {}
    for i in tqdm.trange(surface.n_cells, desc="Building face map", leave=False):
        neighbors = surface.cell_neighbors(i, connections='edges')
        face_map[i] = tuple(neighbors)
    return face_map


def build_edge_map_2(surface):
    edge_map = {}
    for i in tqdm.trange(surface.n_cells, desc="Building edge map", leave=False):
        cell = surface.get_cell(i)
        for j in range(cell.n_edges):
            edge = tuple(sorted(cell.edges[j].point_ids))
            if edge not in edge_map.keys():
                edge_map[edge] = [i]
            else:
                edge_map[edge].append(i)
    face_map = {}
    for i in range(surface.n_cells):
        neighbors = []
        cell = surface.get_cell(i)
        for j in range(cell.n_edges):
            edge = tuple(sorted(cell.edges[j].point_ids))
            neighbors.extend(edge_map[edge])
        neighbors = list(set(neighbors))
        neighbors.remove(i)
        face_map[i] = tuple(neighbors)
    return face_map


def build_face_neighbors(mesh):
    """
    Build a list of face neighbors for each face in a PyVista PolyData surface mesh.
    Parameters:
    - mesh (pv.PolyData): Input PyVista PolyData surface mesh.
    Returns:
    - neighbors_list (list of lists): For each face, a list of indices of neighboring faces.
    """
    # Ensure we're working with a clean triangulated mesh
    edges_to_cells = defaultdict(set)
    neighbors_list = [[] for _ in range(mesh.n_cells)]
    # Build the mapping from edges to cells
    for i in range(mesh.n_cells):
        cell_point_ids = mesh.get_cell(i).point_ids
        edges = [(cell_point_ids[j], cell_point_ids[(j + 1) % 3]) for j in range(3)]
        for edge in edges:
            sorted_edge = tuple(sorted(edge))  # Sort the tuple to avoid directional issues
            edges_to_cells[sorted_edge].add(i)
    # Use the mapping to find neighbors
    for edge, cells in edges_to_cells.items():
        for cell_id in cells:
            neighbors_list[cell_id].extend(cells - {cell_id})
    # Remove duplicates and sort (optional, for cleaner output)
    neighbors_list = [sorted(list(set(neighbors))) for neighbors in neighbors_list]
    neighbors_map = {}
    for i in range(len(neighbors_list)):
        neighbors_map[i] = tuple(neighbors_list[i])
    return neighbors_map


def correct_normals(cell_normals, face_vertices, mesh):
    avg = numpy.mean(mesh.point_normals[face_vertices, :], axis=1)
    for i in tqdm.trange(cell_normals.shape[0], desc="Correcting cell normals", leave=False):
        if numpy.dot(avg[i, :], cell_normals[i, :]) < 0:
            cell_normals[i, :] *= -1
    return cell_normals


def partition(unallocated, face_vertices, cell_normals, angle, vertex_map, edge_neighbors, collapsed):
    faces = []
    allocated = set()
    collapse_check = {}
    #os.mkdir('partition_algorithm')
    while len(unallocated) > 0:
        current_element = unallocated.pop()
        #if collapsed[current_element]:
        #    unallocated.update([current_element])
        #    continue
        allocated.update([current_element])
        check = set(tuple([current_element]))
        current_face = []
        while len(check) > 0:
            current_id = check.pop()
            #current_vertices = face_vertices[current_id, :]
            current_normal = cell_normals[current_id, :]
            #neighbors = []
            #for i in range(face_vertices.shape[1]):
            #    neighbors.extend(vertex_map[current_vertices[i]])
            neighbors = edge_neighbors[current_id]
            #neighbors = set(tuple(neighbors))
            neighbors = set(neighbors)
            #neighbors.remove(current_id)
            neighbors = neighbors.difference(allocated)
            neighbors = neighbors.difference(check)
            for i in neighbors:
                face_angle = numpy.arccos(numpy.clip(numpy.dot(current_normal, cell_normals[i, :]), -1, 1))
                face_angle = (face_angle/numpy.pi)*180.0
                if face_angle < angle:
                    check.update([i])
                #elif face_angle > angle and collapsed[i]:
                #    unallocated.remove(i)
                #    allocated.update([i])
                #    current_face.append(i)
            if current_id in unallocated:
                unallocated.remove(current_id)
            allocated.update([current_id])
            current_face.append(current_id)
            #if len(faces) == 0:
            #    plotter = pyvista.Plotter()
            #    plotter.add_mesh(mesh, color='grey', opacity=0.15)
            #    accepted_cells = mesh.extract_cells(current_face)
            #    check_cells = mesh.extract_cells(list(check))
            #    neighbor_cells = mesh.extract_cells(list(neighbors))
            #    plotter.add_mesh(accepted_cells, color='green')
            #    plotter.add_mesh(check_cells, color='yellow')
            #    plotter.add_mesh(neighbor_cells, color='red')
            #    plotter.show()
        faces.append(current_face)
    return faces



# cython: language_level=3
# distutils: language=c++
import numpy as np
cimport numpy as cnp
import cython
from libc.math cimport sqrt, pow, M_PI
from libcpp.vector cimport vector
from libc.math cimport HUGE_VAL
from libcpp.unordered_map cimport unordered_map
from libcpp.string cimport string


cnp.import_array()

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline double sqr(double x) nogil:
    return x * x


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline double length_3d(double[:] p0, double[:] p1) nogil:
    cdef double dx = p0[0] - p1[0]
    cdef double dy = p0[1] - p1[1]
    cdef double dz = p0[2] - p1[2]
    return sqrt(dx*dx + dy*dy + dz*dz)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cdef inline double norm3(double dx, double dy, double dz) nogil:
    return sqrt(dx*dx + dy*dy + dz*dz)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef norm(vector[double] difference):
    cdef double result = 0.0
    for ii in range(difference.size()):
        result += pow(difference[ii], 2.0)
    return sqrt(result)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef inline void diff(double[:] point_0, double[:] point_1, double[3] result):
    result[0] = point_0[0] - point_1[0]
    result[1] = point_0[1] - point_1[1]
    result[2] = point_0[2] - point_1[2]

@cython.cpow(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef length(double[:] point_0, double[:] point_1, double result):
    result = sqrt(pow(point_0[0] - point_1[0], 2.0) + pow(point_0[1] - point_1[1], 2.0) + pow(point_0[2] - point_1[2], 2.0))

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef basis(double[:] point_0, double[:] point_1):
    cdef vector[double] u_basis = [1.0, 1.0, 1.0]
    cdef vector[double] v_basis = [1.0, 1.0, 1.0]
    cdef vector[double] w_basis = [1.0, 1.0, 1.0]
    cdef double[3] tmp
    cdef double magnitude
    diff(point_1[:], point_0[:], tmp)
    magnitude = norm3(tmp[0], tmp[1], tmp[2])
    w_basis[0] = tmp[0] / magnitude
    w_basis[1] = tmp[1] / magnitude
    w_basis[2] = tmp[2] / magnitude
    if w_basis[2] == -1.0:
        u_basis[0] = -1.0
        v_basis[1] = -1.0
    else:
        u_basis[0] = 1.0 - pow(w_basis[0], 2.0) / (1.0 + w_basis[2])
        u_basis[1] = (-w_basis[0] * w_basis[1]) / (1.0 + w_basis[2])
        u_basis[2] = -w_basis[0]
        v_basis[0] = (-w_basis[0] * w_basis[1]) / (1.0 + w_basis[2])
        v_basis[1] = 1.0 - pow(w_basis[1], 2.0) / (1.0 + w_basis[2])
        v_basis[2] = -w_basis[1]
    return u_basis, v_basis, w_basis

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def basis3(double[:] point_0, double[:] point_1):
    """
    Given two 3D points, returns (u_basis, v_basis, w_basis),
    each a 3-element Python list describing orthonormal directions.
    """
    #cdef double dx = point_1[0] - point_0[0]
    #cdef double dy = point_1[1] - point_0[1]
    #cdef double dz = point_1[2] - point_0[2]
    cdef double dx
    cdef double dy
    cdef double dz
    cdef double[3] difference
    diff(point_0, point_1, difference)
    dx = difference[0]
    dy = difference[1]
    dz = difference[2]
    cdef double mag = norm3(difference[0], difference[1], difference[2])
    cdef double denom
    # Allocate local arrays for u, v, w
    cdef double u[3]
    cdef double v[3]
    cdef double w[3]

    # w = (dx, dy, dz) normalized
    w[0] = dx / mag
    w[1] = dy / mag
    w[2] = dz / mag

    if w[2] == -1.0:
        # The "special case" in your logic
        u[0] = -1.0
        u[1] =  0.0
        u[2] =  0.0
        v[0] =  0.0
        v[1] = -1.0
        v[2] =  0.0
    else:
        # Normal case
        # u_basis[0] = 1.0 - (w_basis[0]^2 / (1.0 + w_basis[2]))
        # etc...
        # We'll do it directly with multiplications:
        denom = 1.0 + w[2]

        u[0] = 1.0 - (w[0]*w[0] / denom)
        u[1] = -(w[0]*w[1]) / denom
        u[2] = -w[0]

        v[0] = -(w[0]*w[1]) / denom
        v[1] = 1.0 - (w[1]*w[1] / denom)
        v[2] = -w[1]

    # Convert to Python objects
    # (If performance is critical and you only use them in Cython,
    #  consider returning them as memoryviews or c arrays.)
    return ([u[0], u[1], u[2]],
            [v[0], v[1], v[2]],
            [w[0], w[1], w[2]])


@cython.cpow(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def bifurcation_cost(double[:] point, double[:,:] data, double[:] terminal, long[:,:] idxs,
                     int vessel, double gamma, double nu, double terminal_flow, double terminal_pressure,
                     double root_pressure, double radius_exponent, double length_exponent):
    cdef vector[double] reduced_resistances
    cdef vector[double] reduced_lengths
    cdef vector[double] bifurcations
    cdef vector[double] flows
    cdef vector[int] main_idx
    cdef vector[int] alt_idx
    cdef vector[double] new_scale # this replaces the data at column index 28 along the main_idx
    cdef vector[double] alt_scale # this replaces the data at column index 28 along the alt_idx
    cdef double upstream_length, downstream_length, terminal_length, r_terminal,\
                r_terminal_sister, f_terminal, f_terminal_sister, r0, l0, f_changed, f_stagnant
    cdef int previous, position, alt
    cdef double[3] upstream_diff, downstream_diff, terminal_diff
    diff(point, data[vessel, 0:3], upstream_diff)
    diff(point, data[vessel, 3:6], downstream_diff)
    diff(point, terminal, terminal_diff)
    upstream_length = norm3(upstream_diff[0], upstream_diff[1], upstream_diff[2])
    downstream_length = norm3(downstream_diff[0], downstream_diff[1], downstream_diff[2])
    terminal_length = norm3(terminal_diff[0], terminal_diff[1], terminal_diff[2])
    r_terminal = ((8.0 * nu) / M_PI) * terminal_length
    r_terminal_sister = ((8.0 * nu) / M_PI) * downstream_length + (data[vessel, 25] - ((8.0 * nu) / M_PI) * data[vessel, 20])
    f_terminal = pow(1+pow(((data[vessel, 22] * r_terminal_sister) / (terminal_flow * r_terminal)), gamma/4.0), -1.0/gamma)
    f_terminal_sister = pow(1+pow(((data[vessel, 22]*r_terminal_sister) / (terminal_flow * r_terminal)), -gamma/4.0), -1.0/gamma)
    r0 = ((8.0 * nu) / M_PI) * upstream_length + pow(((pow(f_terminal, 4.0))/r_terminal +
                                                     (pow(f_terminal_sister, 4.0))/r_terminal_sister), -1.0)
    l0 = pow(f_terminal, radius_exponent) * pow(terminal_length, length_exponent) + \
         pow(f_terminal_sister, radius_exponent) * (pow(downstream_length, length_exponent) + data[vessel, 27])
    reduced_resistances.push_back(r0)
    reduced_lengths.push_back(l0)
    flows.push_back(data[vessel, 22] + terminal_flow)
    main_idx.push_back(vessel)
    bifurcations.push_back(f_terminal_sister)
    bifurcations.push_back(f_terminal)
    if vessel == 0:
        l0 = pow(upstream_length, length_exponent) + l0
        r0 = pow((((data[vessel, 22] + terminal_flow) * r0) / (root_pressure - terminal_pressure)), (1.0/4.0))
        alt_idx.push_back(-1)
        new_scale.push_back(1)
        alt_scale.push_back(1)
        return r0, l0, f_terminal, f_terminal_sister, 1.0, point, reduced_resistances,\
               reduced_lengths, bifurcations, flows, main_idx, alt_idx, new_scale, alt_scale,\
               r_terminal, r_terminal_sister
    previous = vessel
    vessel = idxs[vessel, 2]
    if idxs[vessel, 0] == previous:
        alt = idxs[vessel, 1]
        if alt == -1:
            alt = -2
        alt_idx.push_back(alt)
        position = 0
    else:
        alt = idxs[vessel, 0]
        if alt == -1:
            alt = -2
        alt_idx.push_back(alt)
        position = 1
    if alt == -2:
        f_changed = 1.0
        f_stagnant = 0.0
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister * f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
        l0 = pow(f_changed, radius_exponent) * (pow(upstream_length, length_exponent) + l0)
    else:
        f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)), (gamma/4.0))), (-1.0/gamma))
        f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)), (-gamma/4.0))), (-1.0/gamma))
        f_terminal = f_terminal*f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + pow(((pow(f_changed, 4.0)) /r0 + (pow(f_stagnant, 4.0)) /data[alt, 25]), -1.0)
        l0 = pow(f_changed, radius_exponent) *(pow(upstream_length, length_exponent) + l0) + \
             pow(f_stagnant, radius_exponent) * (pow(data[alt, 20], length_exponent) + data[alt, 27])
    reduced_resistances.push_back(r0)
    reduced_lengths.push_back(l0)
    flows.push_back(data[vessel, 22] + terminal_flow)
    main_idx.push_back(vessel)
    new_scale.push_back(f_changed)
    alt_scale.push_back(f_stagnant)
    if position == 0:
        bifurcations.push_back(f_changed)
        bifurcations.push_back(f_stagnant)
    else:
        bifurcations.push_back(f_stagnant)
        bifurcations.push_back(f_changed)
    if vessel == 0:
        l0 = pow(data[vessel, 20], length_exponent) + l0
        r0 = pow((((data[vessel, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
        return r0, l0, f_terminal,f_terminal_sister,f_parent,point,reduced_resistances,\
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,r_terminal,\
                r_terminal_sister
    previous = vessel
    vessel = idxs[vessel, 2]
    while vessel >= 0:
        if idxs[vessel, 0] == previous:
            alt = idxs[vessel, 1]
            if alt == -1:
                alt = -2
            alt_idx.push_back(alt)
            position = 0
        else:
            alt = idxs[vessel, 0]
            if alt == -1:
                alt = -2
            alt_idx.push_back(alt)
            position = 1
        if alt == -2:
            f_changed = 1.0
            f_stagnant = 0.0
            r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
            l0 = pow(f_changed, length_exponent) * (data[previous, 20] + l0)
            f_terminal = f_terminal * f_changed
            f_terminal_sister = f_terminal_sister * f_changed
            f_parent = f_parent * f_changed
        else:
            f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((terminal_flow+data[previous, 22])*r0)), (gamma/4.0))), (-1.0/gamma))
            f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt,25])/((terminal_flow+data[previous,22])*r0)), (-gamma/4.0))), (-1.0/gamma))
            f_terminal = f_terminal*f_changed
            f_terminal_sister = f_terminal_sister*f_changed
            f_parent = f_parent*f_changed
            r0 = ((8.0 * nu) / M_PI) * data[vessel,20] + pow(((pow(f_changed, 4.0) / r0) + (pow(f_stagnant, 4.0) / data[alt, 25])), -1.0)
            l0 = pow(f_changed, radius_exponent) * (pow(data[previous, 20], length_exponent) + l0) + \
                 pow(f_stagnant, radius_exponent) *(pow(data[alt, 20], length_exponent) + data[alt, 27])
        reduced_resistances.push_back(r0)
        reduced_lengths.push_back(l0)
        flows.push_back(data[vessel, 22] + terminal_flow)
        main_idx.push_back(vessel)
        if vessel >= 0:
            for j in range(new_scale.size()):
                new_scale[j] = new_scale[j]*f_changed
                alt_scale[j] = alt_scale[j]*f_stagnant
            new_scale.push_back(f_changed)
            alt_scale.push_back(f_stagnant)
        if position == 0:
            bifurcations.push_back(f_changed)
            bifurcations.push_back(f_stagnant)
        else:
            bifurcations.push_back(f_stagnant)
            bifurcations.push_back(f_changed)
        previous = vessel
        vessel = idxs[vessel, 2]
    l0 = pow(data[0, 20], length_exponent) + l0
    r0 = pow((((data[0, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
    return r0, l0, f_terminal, f_terminal_sister, f_parent, point, reduced_resistances, \
           reduced_lengths, bifurcations, flows, main_idx, alt_idx, new_scale, alt_scale, r_terminal, \
           r_terminal_sister

@cython.cpow(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def tree_cost(double[:] point, double[:,:] data, double[:] terminal, long[:,:] idxs, int vessel, double gamma, double nu,
              double terminal_flow, double terminal_pressure, double root_pressure, double radius_exponent,
              double length_exponent):
    cdef double upstream_length, downstream_length, terminal_length, r_terminal, r_terminal_sister, r_parent
    cdef double f_terminal, f_terminal_sister, r0, l0, f_changed, f_stagnant, f_parent, cost
    cdef int previous, position, alt
    cdef double[3] direction_upstream
    cdef double[3] direction_downstream
    cdef double[3] direction_terminal
    direction_upstream[0] = point[0] - data[vessel, 0]
    direction_upstream[1] = point[1] - data[vessel, 1]
    direction_upstream[2] = point[2] - data[vessel, 2]
    direction_downstream[0] = point[0] - data[vessel, 3]
    direction_downstream[1] = point[1] - data[vessel, 4]
    direction_downstream[2] = point[2] - data[vessel, 5]
    direction_terminal[0] = point[0] - terminal[0]
    direction_terminal[1] = point[1] - terminal[1]
    direction_terminal[2] = point[2] - terminal[2]
    upstream_length = sqrt(direction_upstream[0] ** 2.0 + direction_upstream[1] ** 2.0 + direction_upstream[2] ** 2.0)
    downstream_length = sqrt(direction_downstream[0] ** 2.0 + direction_downstream[1] ** 2.0 + direction_downstream[2] ** 2.0)
    terminal_length = sqrt(direction_terminal[0] ** 2.0 + direction_terminal[1] ** 2.0 + direction_terminal[2] ** 2.0)
    #upstream_length = norm(diff(point, data[vessel, 0:3]))
    #downstream_length = norm(diff(point, data[vessel, 3:6]))
    #terminal_length = norm(diff(point, terminal))
    #upstream_length = 0.0
    #downstream_length = 0.0
    #terminal_length = 0.0
    #length(point, data[vessel, 0:3], upstream_length)
    #length(point, data[vessel, 3:6], downstream_length)
    #length(point, terminal, terminal_length)
    if upstream_length == 0.0:
        return HUGE_VAL
    elif downstream_length == 0.0:
        return HUGE_VAL
    elif terminal_length == 0.0:
        return HUGE_VAL
    r_terminal = ((8.0 * nu) / M_PI) * terminal_length
    r_terminal_sister = ((8.0 * nu) / M_PI) * downstream_length + (data[vessel, 25] -
                                                                   ((8.0 * nu) / M_PI) * data[vessel, 20])
    f_terminal = pow(1+pow(((data[vessel, 22] * r_terminal_sister) / (terminal_flow * r_terminal)), gamma/4.0),
                     -1.0/gamma)
    f_terminal_sister = pow(1+pow(((data[vessel, 22]*r_terminal_sister) / (terminal_flow * r_terminal)), -gamma/4.0),
                            -1.0/gamma)
    f_parent = 1.0
    r0 = ((8.0 * nu) / M_PI) * upstream_length + pow(((pow(f_terminal, 4.0))/r_terminal +
                                                     (pow(f_terminal_sister, 4.0))/r_terminal_sister), -1.0)
    r_parent = r0
    l0 = pow(f_terminal, radius_exponent) * pow(terminal_length, length_exponent) + \
         pow(f_terminal_sister, radius_exponent) * (pow(downstream_length, length_exponent) + data[vessel, 27])
    if vessel == 0:
        l0 = pow(upstream_length, length_exponent) + l0
        r0 = pow((((data[vessel, 22] + terminal_flow) * r0) / (root_pressure - terminal_pressure)), (1.0/4.0))
        cost = M_PI * pow(r0, radius_exponent) * l0
        return cost
    previous = vessel
    vessel = idxs[vessel, 2]
    if idxs[vessel, 0] == previous:
        alt = idxs[vessel, 1]
        if alt == -1:
            alt = -2
        position = 0
    else:
        alt = idxs[vessel, 0]
        if alt == -1:
            alt = -2
        position = 1
    if alt == -2:
        f_changed = 1.0
        f_stagnant = 0.0
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister * f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
        l0 = pow(f_changed, radius_exponent) * (pow(upstream_length, length_exponent) + l0)
    else:
        f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)),
                                 (gamma/4.0))), (-1.0/gamma))
        f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)),
                                  (-gamma/4.0))), (-1.0/gamma))
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + pow(((pow(f_changed, 4.0)) /r0 +
                                                           (pow(f_stagnant, 4.0)) /data[alt, 25]), -1.0)
        l0 = pow(f_changed, radius_exponent) *(pow(upstream_length, length_exponent) + l0) + \
             pow(f_stagnant, radius_exponent) * (pow(data[alt, 20], length_exponent) + data[alt, 27])
    if vessel == 0:
        l0 = pow(data[vessel, 20], length_exponent) + l0
        r0 = pow((((data[vessel, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
        cost = M_PI * pow(r0, radius_exponent) * l0
        return cost
    previous = vessel
    vessel = idxs[vessel, 2]
    while vessel >= 0:
        if idxs[vessel, 0] == previous:
            alt = idxs[vessel, 1]
            if alt == -1:
                alt = -2
            position = 0
        else:
            alt = idxs[vessel, 0]
            if alt == -1:
                alt = -2
            position = 1
        if alt == -2:
            f_changed = 1.0
            f_stagnant = 0.0
            r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
            l0 = pow(f_changed, length_exponent) * (data[previous, 20] + l0)
            f_terminal = f_terminal * f_changed
            f_terminal_sister = f_terminal_sister * f_changed
            f_parent = f_parent * f_changed
        else:
            f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((terminal_flow+data[previous, 22])*r0)),
                                     (gamma/4.0))), (-1.0/gamma))
            f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt,25])/((terminal_flow+data[previous,22])*r0)),
                                      (-gamma/4.0))), (-1.0/gamma))
            f_terminal = f_terminal*f_changed
            f_terminal_sister = f_terminal_sister*f_changed
            f_parent = f_parent*f_changed
            r0 = ((8.0 * nu) / M_PI) * data[vessel,20] + pow(((pow(f_changed, 4.0) / r0) +
                                                              (pow(f_stagnant, 4.0) / data[alt, 25])), -1.0)
            l0 = pow(f_changed, radius_exponent) * (pow(data[previous, 20], length_exponent) + l0) + \
                 pow(f_stagnant, radius_exponent) *(pow(data[alt, 20], length_exponent) + data[alt, 27])
        previous = vessel
        vessel = idxs[vessel, 2]
    l0 = pow(data[0, 20], length_exponent) + l0
    r0 = pow((((data[0, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
    cost = M_PI * pow(r0, radius_exponent) * l0
    return cost

@cython.cpow(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def create_new_vessels(double[:] point, double[:,:] data, double[:] terminal, double[:,:] terminal_vessel,
                       double[:,:] terminal_sister_vessel, double[:,:] terminal_parent_vessel, double max_node,
                       double max_vessel, long[:,:] idxs, int vessel, double gamma, double nu, double terminal_flow,
                       double terminal_pressure, double root_pressure, double radius_exponent, double length_exponent):
    cdef double upstream_length, downstream_length, terminal_length, r_terminal, r_terminal_sister, r_parent
    cdef double f_terminal, f_terminal_sister, r0, l0, f_changed, f_stagnant, f_parent, cost, rl_parent
    cdef double terminal_bifurcation_ratio, terminal_sister_bifurcation_ratio
    cdef int previous, position, alt, original_vessel
    original_vessel = vessel
    cdef double[3] upstream_diff, downstream_diff, terminal_diff
    diff(point, data[vessel, 0:3], upstream_diff)
    diff(point, data[vessel, 3:6], downstream_diff)
    diff(point, terminal, terminal_diff)
    upstream_length = norm3(upstream_diff[0], upstream_diff[1], upstream_diff[2])
    downstream_length = norm3(downstream_diff[0], downstream_diff[1], downstream_diff[2])
    terminal_length = norm3(terminal_diff[0], terminal_diff[1], terminal_diff[2])
    r_terminal = ((8.0 * nu) / M_PI) * terminal_length
    r_terminal_sister = ((8.0 * nu) / M_PI) * downstream_length + (data[vessel, 25] -
                                                                   ((8.0 * nu) / M_PI) * data[vessel, 20])
    f_terminal = pow(1+pow(((data[vessel, 22] * r_terminal_sister) / (terminal_flow * r_terminal)), gamma/4.0),
                     -1.0/gamma)
    terminal_bifurcation_ratio = f_terminal
    f_terminal_sister = pow(1+pow(((data[vessel, 22]*r_terminal_sister) / (terminal_flow * r_terminal)), -gamma/4.0),
                            -1.0/gamma)
    terminal_sister_bifurcation_ratio = f_terminal_sister
    f_parent = 1.0
    r0 = ((8.0 * nu) / M_PI) * upstream_length + pow(((pow(f_terminal, 4.0))/r_terminal +
                                                     (pow(f_terminal_sister, 4.0))/r_terminal_sister), -1.0)
    r_parent = r0
    l0 = pow(f_terminal, radius_exponent) * pow(terminal_length, length_exponent) + \
         pow(f_terminal_sister, radius_exponent) * (pow(downstream_length, length_exponent) + data[vessel, 27])
    rl_parent = l0
    if vessel == 0:
        l0 = pow(upstream_length, length_exponent) + l0
        r0 = pow((((data[vessel, 22] + terminal_flow) * r0) / (root_pressure - terminal_pressure)), (1.0/4.0))
        terminal_vessel[0, 0:3] = point[:]
        terminal_vessel[0, 3:6] = terminal[:]
        u, v, w = basis3(terminal_vessel[0, 0:3], terminal_vessel[0, 3:6])
        terminal_vessel[0, 6] = u[0]
        terminal_vessel[0, 7] = u[1]
        terminal_vessel[0, 8] = u[2]
        terminal_vessel[0, 9] = v[0]
        terminal_vessel[0, 10] = v[1]
        terminal_vessel[0, 11] = v[2]
        terminal_vessel[0, 12] = w[0]
        terminal_vessel[0, 13] = w[1]
        terminal_vessel[0, 14] = w[2]
        terminal_vessel[0, 17] = original_vessel * 1.0
        terminal_vessel[0, 18] = max_node + 1.0
        terminal_vessel[0, 19] = max_node + 2.0
        terminal_vessel[0, 20] = terminal_length
        terminal_vessel[0, 21] = r0 * f_terminal
        terminal_vessel[0, 22] = terminal_flow
        terminal_vessel[0, 25] = r_terminal
        terminal_vessel[0, 26] = data[original_vessel, 26] + 1.0
        terminal_vessel[0, 27] = 0.0
        terminal_vessel[0, 28] = f_terminal
        terminal_sister_vessel[0, 0:3] = point[:]
        terminal_sister_vessel[0, 3:6] = data[original_vessel, 3:6]
        u, v, w = basis3(terminal_sister_vessel[0, 0:3], terminal_sister_vessel[0, 3:6])
        terminal_sister_vessel[0, 6] = u[0]
        terminal_sister_vessel[0, 7] = u[1]
        terminal_sister_vessel[0, 8] = u[2]
        terminal_sister_vessel[0, 9] = v[0]
        terminal_sister_vessel[0, 10] = v[1]
        terminal_sister_vessel[0, 11] = v[2]
        terminal_sister_vessel[0, 12] = w[0]
        terminal_sister_vessel[0, 13] = w[1]
        terminal_sister_vessel[0, 14] = w[2]
        terminal_sister_vessel[0, 15] = data[original_vessel, 15]
        terminal_sister_vessel[0, 16] = data[original_vessel, 16]
        terminal_sister_vessel[0, 17] = original_vessel * 1.0
        terminal_sister_vessel[0, 18] = max_node + 1.0
        terminal_sister_vessel[0, 19] = data[original_vessel, 19]
        terminal_sister_vessel[0, 20] = downstream_length
        terminal_sister_vessel[0, 21] = r0 * f_terminal_sister
        terminal_sister_vessel[0, 22] = data[original_vessel, 22]
        terminal_sister_vessel[0, 23] = data[original_vessel, 23]
        terminal_sister_vessel[0, 24] = data[original_vessel, 24]
        terminal_sister_vessel[0, 25] = r_terminal_sister
        terminal_sister_vessel[0, 26] = data[original_vessel, 26] + 1.0
        terminal_sister_vessel[0, 27] = data[original_vessel, 27]
        terminal_sister_vessel[0, 28] = f_terminal_sister
        terminal_parent_vessel[0, 0:3] = data[original_vessel, 0:3]
        terminal_parent_vessel[0, 3:6] = point[:]
        u, v, w = basis3(terminal_parent_vessel[0, 0:3], terminal_parent_vessel[0, 3:6])
        terminal_parent_vessel[0, 6] = u[0]
        terminal_parent_vessel[0, 7] = u[1]
        terminal_parent_vessel[0, 8] = u[2]
        terminal_parent_vessel[0, 9] = v[0]
        terminal_parent_vessel[0, 10] = v[1]
        terminal_parent_vessel[0, 11] = v[2]
        terminal_parent_vessel[0, 12] = w[0]
        terminal_parent_vessel[0, 13] = w[1]
        terminal_parent_vessel[0, 14] = w[2]
        terminal_parent_vessel[0, 15] = max_vessel
        terminal_parent_vessel[0, 16] = max_vessel + 1.0
        terminal_parent_vessel[0, 17] = data[original_vessel, 17]
        terminal_parent_vessel[0, 18] = data[original_vessel, 18]
        terminal_parent_vessel[0, 19] = max_node + 1.0
        terminal_parent_vessel[0, 20] = upstream_length
        terminal_parent_vessel[0, 21] = r0 * f_parent
        terminal_parent_vessel[0, 22] = data[original_vessel, 22] + terminal_flow
        terminal_parent_vessel[0, 23] = terminal_bifurcation_ratio
        terminal_parent_vessel[0, 24] = terminal_sister_bifurcation_ratio
        terminal_parent_vessel[0, 25] = r_parent
        terminal_parent_vessel[0, 26] = data[original_vessel, 26]
        terminal_parent_vessel[0, 27] = rl_parent
        terminal_parent_vessel[0, 28] = f_parent
        return
    previous = vessel
    vessel = idxs[vessel, 2]
    if idxs[vessel, 0] == previous:
        alt = idxs[vessel, 1]
        if alt == -1:
            alt = -2
        position = 0
    else:
        alt = idxs[vessel, 0]
        if alt == -1:
            alt = -2
        position = 1
    if alt == -2:
        f_changed = 1.0
        f_stagnant = 0.0
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister * f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
        l0 = pow(f_changed, radius_exponent) * (pow(upstream_length, length_exponent) + l0)
    else:
        f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)),
                                 (gamma/4.0))), (-1.0/gamma))
        f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)),
                                  (-gamma/4.0))), (-1.0/gamma))
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister * f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + pow(((pow(f_changed, 4.0)) /r0 +
                                                           (pow(f_stagnant, 4.0)) /data[alt, 25]), -1.0)
        l0 = pow(f_changed, radius_exponent) *(pow(upstream_length, length_exponent) + l0) + \
             pow(f_stagnant, radius_exponent) * (pow(data[alt, 20], length_exponent) + data[alt, 27])
    if vessel == 0:
        l0 = pow(data[vessel, 20], length_exponent) + l0
        r0 = pow((((data[vessel, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
        terminal_vessel[0, 0:3] = point[:]
        terminal_vessel[0, 3:6] = terminal[:]
        u, v, w = basis3(terminal_vessel[0, 0:3], terminal_vessel[0, 3:6])
        terminal_vessel[0, 6] = u[0]
        terminal_vessel[0, 7] = u[1]
        terminal_vessel[0, 8] = u[2]
        terminal_vessel[0, 9] = v[0]
        terminal_vessel[0, 10] = v[1]
        terminal_vessel[0, 11] = v[2]
        terminal_vessel[0, 12] = w[0]
        terminal_vessel[0, 13] = w[1]
        terminal_vessel[0, 14] = w[2]
        terminal_vessel[0, 17] = original_vessel * 1.0
        terminal_vessel[0, 18] = max_node + 1.0
        terminal_vessel[0, 19] = max_node + 2.0
        terminal_vessel[0, 20] = terminal_length
        terminal_vessel[0, 21] = r0 * f_terminal
        terminal_vessel[0, 22] = terminal_flow
        terminal_vessel[0, 25] = r_terminal
        terminal_vessel[0, 26] = data[original_vessel, 26] + 1.0
        terminal_vessel[0, 27] = 0.0
        terminal_vessel[0, 28] = f_terminal
        terminal_sister_vessel[0, 0:3] = point[:]
        terminal_sister_vessel[0, 3:6] = data[original_vessel, 3:6]
        u, v, w = basis3(terminal_sister_vessel[0, 0:3], terminal_sister_vessel[0, 3:6])
        terminal_sister_vessel[0, 6] = u[0]
        terminal_sister_vessel[0, 7] = u[1]
        terminal_sister_vessel[0, 8] = u[2]
        terminal_sister_vessel[0, 9] = v[0]
        terminal_sister_vessel[0, 10] = v[1]
        terminal_sister_vessel[0, 11] = v[2]
        terminal_sister_vessel[0, 12] = w[0]
        terminal_sister_vessel[0, 13] = w[1]
        terminal_sister_vessel[0, 14] = w[2]
        terminal_sister_vessel[0, 15] = data[original_vessel, 15]
        terminal_sister_vessel[0, 16] = data[original_vessel, 16]
        terminal_sister_vessel[0, 17] = original_vessel * 1.0
        terminal_sister_vessel[0, 18] = max_node + 1.0
        terminal_sister_vessel[0, 19] = data[original_vessel, 19]
        terminal_sister_vessel[0, 20] = downstream_length
        terminal_sister_vessel[0, 21] = r0 * f_terminal_sister
        terminal_sister_vessel[0, 22] = data[original_vessel, 22]
        terminal_sister_vessel[0, 23] = data[original_vessel, 23]
        terminal_sister_vessel[0, 24] = data[original_vessel, 24]
        terminal_sister_vessel[0, 25] = r_terminal_sister
        terminal_sister_vessel[0, 26] = data[original_vessel, 26] + 1.0
        terminal_sister_vessel[0, 27] = data[original_vessel, 27]
        terminal_sister_vessel[0, 28] = f_terminal_sister
        terminal_parent_vessel[0, 0:3] = data[original_vessel, 0:3]
        terminal_parent_vessel[0, 3:6] = point[:]
        u, v, w = basis3(terminal_parent_vessel[0, 0:3], terminal_parent_vessel[0, 3:6])
        terminal_parent_vessel[0, 6] = u[0]
        terminal_parent_vessel[0, 7] = u[1]
        terminal_parent_vessel[0, 8] = u[2]
        terminal_parent_vessel[0, 9] = v[0]
        terminal_parent_vessel[0, 10] = v[1]
        terminal_parent_vessel[0, 11] = v[2]
        terminal_parent_vessel[0, 12] = w[0]
        terminal_parent_vessel[0, 13] = w[1]
        terminal_parent_vessel[0, 14] = w[2]
        terminal_parent_vessel[0, 15] = max_vessel
        terminal_parent_vessel[0, 16] = max_vessel + 1.0
        terminal_parent_vessel[0, 17] = data[original_vessel, 17]
        terminal_parent_vessel[0, 18] = data[original_vessel, 18]
        terminal_parent_vessel[0, 19] = max_node + 1.0
        terminal_parent_vessel[0, 20] = upstream_length
        terminal_parent_vessel[0, 21] = r0 * f_parent
        terminal_parent_vessel[0, 22] = data[original_vessel, 22] + terminal_flow
        terminal_parent_vessel[0, 23] = terminal_bifurcation_ratio
        terminal_parent_vessel[0, 24] = terminal_sister_bifurcation_ratio
        terminal_parent_vessel[0, 25] = r_parent
        terminal_parent_vessel[0, 26] = data[original_vessel, 26]
        terminal_parent_vessel[0, 27] = rl_parent
        terminal_parent_vessel[0, 28] = f_parent
        return
    previous = vessel
    vessel = idxs[vessel, 2]
    while vessel >= 0:
        if idxs[vessel, 0] == previous:
            alt = idxs[vessel, 1]
            if alt == -1:
                alt = -2
            position = 0
        else:
            alt = idxs[vessel, 0]
            if alt == -1:
                alt = -2
            position = 1
        if alt == -2:
            f_changed = 1.0
            f_stagnant = 0.0
            r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
            l0 = pow(f_changed, length_exponent) * (data[previous, 20] + l0)
            f_terminal = f_terminal * f_changed
            f_terminal_sister = f_terminal_sister * f_changed
            f_parent = f_parent * f_changed
        else:
            f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((terminal_flow+data[previous, 22])*r0)),
                                     (gamma/4.0))), (-1.0/gamma))
            f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt,25])/((terminal_flow+data[previous,22])*r0)),
                                      (-gamma/4.0))), (-1.0/gamma))
            f_terminal = f_terminal * f_changed
            f_terminal_sister = f_terminal_sister * f_changed
            f_parent = f_parent * f_changed
            r0 = ((8.0 * nu) / M_PI) * data[vessel,20] + pow(((pow(f_changed, 4.0) / r0) +
                                                              (pow(f_stagnant, 4.0) / data[alt, 25])), -1.0)
            l0 = pow(f_changed, radius_exponent) * (pow(data[previous, 20], length_exponent) + l0) + \
                 pow(f_stagnant, radius_exponent) *(pow(data[alt, 20], length_exponent) + data[alt, 27])
        previous = vessel
        vessel = idxs[vessel, 2]
    l0 = pow(data[0, 20], length_exponent) + l0
    r0 = pow((((data[0, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
    terminal_vessel[0, 0:3] = point[:]
    terminal_vessel[0, 3:6] = terminal[:]
    u, v, w = basis3(terminal_vessel[0, 0:3], terminal_vessel[0, 3:6])
    terminal_vessel[0, 6] = u[0]
    terminal_vessel[0, 7] = u[1]
    terminal_vessel[0, 8] = u[2]
    terminal_vessel[0, 9] = v[0]
    terminal_vessel[0, 10] = v[1]
    terminal_vessel[0, 11] = v[2]
    terminal_vessel[0, 12] = w[0]
    terminal_vessel[0, 13] = w[1]
    terminal_vessel[0, 14] = w[2]
    terminal_vessel[0, 17] = original_vessel * 1.0
    terminal_vessel[0, 18] = max_node + 1.0
    terminal_vessel[0, 19] = max_node + 2.0
    terminal_vessel[0, 20] = terminal_length
    terminal_vessel[0, 21] = r0 * f_terminal
    terminal_vessel[0, 22] = terminal_flow
    terminal_vessel[0, 25] = r_terminal
    terminal_vessel[0, 26] = data[original_vessel, 26] + 1.0
    terminal_vessel[0, 27] = 0.0
    terminal_vessel[0, 28] = f_terminal
    terminal_sister_vessel[0, 0:3] = point[:]
    terminal_sister_vessel[0, 3:6] = data[original_vessel, 3:6]
    u, v, w = basis3(terminal_sister_vessel[0, 0:3], terminal_sister_vessel[0, 3:6])
    terminal_sister_vessel[0, 6] = u[0]
    terminal_sister_vessel[0, 7] = u[1]
    terminal_sister_vessel[0, 8] = u[2]
    terminal_sister_vessel[0, 9] = v[0]
    terminal_sister_vessel[0, 10] = v[1]
    terminal_sister_vessel[0, 11] = v[2]
    terminal_sister_vessel[0, 12] = w[0]
    terminal_sister_vessel[0, 13] = w[1]
    terminal_sister_vessel[0, 14] = w[2]
    terminal_sister_vessel[0, 15] = data[original_vessel, 15]
    terminal_sister_vessel[0, 16] = data[original_vessel, 16]
    terminal_sister_vessel[0, 17] = original_vessel * 1.0
    terminal_sister_vessel[0, 18] = max_node + 1.0
    terminal_sister_vessel[0, 19] = data[original_vessel, 19]
    terminal_sister_vessel[0, 20] = downstream_length
    terminal_sister_vessel[0, 21] = r0 * f_terminal_sister
    terminal_sister_vessel[0, 22] = data[original_vessel, 22]
    terminal_sister_vessel[0, 23] = data[original_vessel, 23]
    terminal_sister_vessel[0, 24] = data[original_vessel, 24]
    terminal_sister_vessel[0, 25] = r_terminal_sister
    terminal_sister_vessel[0, 26] = data[original_vessel, 26] + 1.0
    terminal_sister_vessel[0, 27] = data[original_vessel, 27]
    terminal_sister_vessel[0, 28] = f_terminal_sister
    terminal_parent_vessel[0, 0:3] = data[original_vessel, 0:3]
    terminal_parent_vessel[0, 3:6] = point[:]
    u, v, w = basis3(terminal_parent_vessel[0, 0:3], terminal_parent_vessel[0, 3:6])
    terminal_parent_vessel[0, 6] = u[0]
    terminal_parent_vessel[0, 7] = u[1]
    terminal_parent_vessel[0, 8] = u[2]
    terminal_parent_vessel[0, 9] = v[0]
    terminal_parent_vessel[0, 10] = v[1]
    terminal_parent_vessel[0, 11] = v[2]
    terminal_parent_vessel[0, 12] = w[0]
    terminal_parent_vessel[0, 13] = w[1]
    terminal_parent_vessel[0, 14] = w[2]
    terminal_parent_vessel[0, 15] = max_vessel
    terminal_parent_vessel[0, 16] = max_vessel + 1.0
    terminal_parent_vessel[0, 17] = data[original_vessel, 17]
    terminal_parent_vessel[0, 18] = data[original_vessel, 18]
    terminal_parent_vessel[0, 19] = max_node + 1.0
    terminal_parent_vessel[0, 20] = upstream_length
    terminal_parent_vessel[0, 21] = r0 * f_parent
    terminal_parent_vessel[0, 22] = data[original_vessel, 22] + terminal_flow
    terminal_parent_vessel[0, 23] = terminal_bifurcation_ratio
    terminal_parent_vessel[0, 24] = terminal_sister_bifurcation_ratio
    terminal_parent_vessel[0, 25] = r_parent
    terminal_parent_vessel[0, 26] = data[original_vessel, 26]
    terminal_parent_vessel[0, 27] = rl_parent
    terminal_parent_vessel[0, 28] = f_parent
    return

@cython.cpow(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def update_vessels(double[:] point, double[:,:] data, double[:] terminal, long[:,:] idxs,
                   int vessel, double gamma, double nu, double terminal_flow, double terminal_pressure,
                   double root_pressure, double radius_exponent, double length_exponent):
    cdef double upstream_length, downstream_length, terminal_length, r_terminal, r_terminal_sister, r_parent
    cdef double f_terminal, f_terminal_sister, r0, l0, f_changed, f_stagnant, f_parent, cost
    cdef vector[vector[double]] bifurcation_ratios
    cdef vector[double] tmp_bifurcation_ratios = [0.0, 0.0]
    cdef vector[int] main_idx
    cdef vector[int] alt_idx
    cdef vector[double] main_scale
    cdef vector[double] alt_scale
    cdef vector[double] reduced_resistance
    cdef vector[double] reduced_length
    cdef vector[double] flows
    cdef int previous, position, alt, alt_position
    cdef double[3] upstream_diff, downstream_diff, terminal_diff
    diff(point, data[vessel, 0:3], upstream_diff)
    diff(point, data[vessel, 3:6], downstream_diff)
    diff(point, terminal, terminal_diff)
    upstream_length = norm3(upstream_diff[0], upstream_diff[1], upstream_diff[2])
    downstream_length = norm3(downstream_diff[0], downstream_diff[1], downstream_diff[2])
    terminal_length = norm3(terminal_diff[0], terminal_diff[1], terminal_diff[2])
    r_terminal = ((8.0 * nu) / M_PI) * terminal_length
    r_terminal_sister = ((8.0 * nu) / M_PI) * downstream_length + (data[vessel, 25] -
                                                                   ((8.0 * nu) / M_PI) * data[vessel, 20])
    f_terminal = pow(1+pow(((data[vessel, 22] * r_terminal_sister) / (terminal_flow * r_terminal)), gamma/4.0),
                     -1.0/gamma)
    f_terminal_sister = pow(1+pow(((data[vessel, 22]*r_terminal_sister) / (terminal_flow * r_terminal)), -gamma/4.0),
                            -1.0/gamma)
    f_parent = 1.0
    r0 = ((8.0 * nu) / M_PI) * upstream_length + pow(((pow(f_terminal, 4.0))/r_terminal +
                                                     (pow(f_terminal_sister, 4.0))/r_terminal_sister), -1.0)
    r_parent = r0
    l0 = pow(f_terminal, radius_exponent) * pow(terminal_length, length_exponent) + \
         pow(f_terminal_sister, radius_exponent) * (pow(downstream_length, length_exponent) + data[vessel, 27])
    if vessel == 0:
        l0 = pow(upstream_length, length_exponent) + l0
        r0 = pow((((data[vessel, 22] + terminal_flow) * r0) / (root_pressure - terminal_pressure)), (1.0/4.0))
        return reduced_resistance, reduced_length, main_idx, main_scale, alt_idx, alt_scale, bifurcation_ratios, flows, r0, l0
    previous = vessel
    vessel = idxs[vessel, 2]
    if idxs[vessel, 0] == previous:
        alt = idxs[vessel, 1]
        if alt == -1:
            alt = -2
        position = 0
        alt_position = 1
    else:
        alt = idxs[vessel, 0]
        if alt == -1:
            alt = -2
        position = 1
        alt_position = 0
    if alt == -2:
        f_changed = 1.0
        f_stagnant = 0.0
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister * f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
        l0 = pow(f_changed, radius_exponent) * (pow(upstream_length, length_exponent) + l0)
    else:
        f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)),
                                 (gamma/4.0))), (-1.0/gamma))
        f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)),
                                  (-gamma/4.0))), (-1.0/gamma))
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + pow(((pow(f_changed, 4.0)) /r0 +
                                                           (pow(f_stagnant, 4.0)) /data[alt, 25]), -1.0)
        l0 = pow(f_changed, radius_exponent) *(pow(upstream_length, length_exponent) + l0) + \
             pow(f_stagnant, radius_exponent) * (pow(data[alt, 20], length_exponent) + data[alt, 27])
    tmp_bifurcation_ratios[position] = f_changed
    tmp_bifurcation_ratios[alt_position] = f_stagnant
    if vessel == 0:
        main_idx.push_back(vessel)
        main_scale.push_back(1.0)
        alt_idx.push_back(alt)
        alt_scale.push_back(f_stagnant)
        reduced_resistance.push_back(r0)
        reduced_length.push_back(l0)
        tmp_bifurcation_ratios[position] = f_changed
        tmp_bifurcation_ratios[alt_position] = f_stagnant
        bifurcation_ratios.push_back(tmp_bifurcation_ratios)
        flows.push_back(data[vessel, 22] + terminal_flow)
        l0 = pow(data[vessel, 20], length_exponent) + l0
        r0 = pow((((data[vessel, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
        return reduced_resistance, reduced_length, main_idx, main_scale, alt_idx, alt_scale, bifurcation_ratios, flows, r0, l0
    else:
        main_idx.push_back(vessel)
        alt_idx.push_back(alt)
        main_scale.push_back(1.0)
        alt_scale.push_back(f_stagnant)
        reduced_resistance.push_back(r0)
        reduced_length.push_back(l0)
        tmp_bifurcation_ratios[position] = f_changed
        tmp_bifurcation_ratios[alt_position] = f_stagnant
        bifurcation_ratios.push_back(tmp_bifurcation_ratios)
        flows.push_back(data[vessel, 22] + terminal_flow)
    # All vessels past this point are above the triad
    previous = vessel
    vessel = idxs[vessel, 2]
    while vessel >= 0:
        if idxs[vessel, 0] == previous:
            alt = idxs[vessel, 1]
            if alt == -1:
                alt = -2
            position = 0
            alt_position = 1
        else:
            alt = idxs[vessel, 0]
            if alt == -1:
                alt = -2
            position = 1
            alt_position = 0
        if alt == -2:
            f_changed = 1.0
            f_stagnant = 0.0
            r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
            l0 = pow(f_changed, length_exponent) * (data[previous, 20] + l0)
            f_terminal = f_terminal * f_changed
            f_terminal_sister = f_terminal_sister * f_changed
            f_parent = f_parent * f_changed
        else:
            f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((terminal_flow+data[previous, 22])*r0)),
                                     (gamma/4.0))), (-1.0/gamma))
            f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt,25])/((terminal_flow+data[previous,22])*r0)),
                                      (-gamma/4.0))), (-1.0/gamma))
            f_terminal = f_terminal * f_changed
            f_terminal_sister = f_terminal_sister * f_changed
            f_parent = f_parent * f_changed
            r0 = ((8.0 * nu) / M_PI) * data[vessel,20] + pow(((pow(f_changed, 4.0) / r0) +
                                                              (pow(f_stagnant, 4.0) / data[alt, 25])), -1.0)
            l0 = pow(f_changed, radius_exponent) * (pow(data[previous, 20], length_exponent) + l0) + \
                 pow(f_stagnant, radius_exponent) *(pow(data[alt, 20], length_exponent) + data[alt, 27])
        for j in range(main_scale.size()):
            main_scale[j] = main_scale[j] * f_changed
            alt_scale[j] = alt_scale[j] * f_changed
        main_idx.push_back(vessel)
        alt_idx.push_back(alt)
        main_scale.push_back(1.0)
        alt_scale.push_back(f_stagnant)
        reduced_resistance.push_back(r0)
        reduced_length.push_back(l0)
        tmp_bifurcation_ratios[position] = f_changed
        tmp_bifurcation_ratios[alt_position] = f_stagnant
        bifurcation_ratios.push_back(tmp_bifurcation_ratios)
        flows.push_back(data[vessel, 22] + terminal_flow)
        previous = vessel
        vessel = idxs[vessel, 2]
    l0 = pow(data[0, 20], length_exponent) + l0
    r0 = pow((((data[0, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
    return reduced_resistance, reduced_length, main_idx, main_scale, alt_idx, alt_scale, bifurcation_ratios, flows, r0, l0

@cython.cpow(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def tree_cost_2(double[:] point, double[:,:] data, double[:] terminal, long[:,:] idxs, int vessel, double gamma, double nu,
              double terminal_flow, double terminal_pressure, double root_pressure, double radius_exponent,
              double length_exponent):
    cdef double upstream_length, downstream_length, terminal_length, r_terminal, r_terminal_sister, r_parent
    cdef double f_terminal, f_terminal_sister, r0, l0, f_changed, f_stagnant, f_parent, cost
    cdef int previous, position, alt
    cdef double[3] direction_upstream
    cdef double[3] direction_downstream
    cdef double[3] direction_terminal
    direction_upstream[0] = point[0] - data[vessel, 0]
    direction_upstream[1] = point[1] - data[vessel, 1]
    direction_upstream[2] = point[2] - data[vessel, 2]
    direction_downstream[0] = point[0] - data[vessel, 3]
    direction_downstream[1] = point[1] - data[vessel, 4]
    direction_downstream[2] = point[2] - data[vessel, 5]
    direction_terminal[0] = point[0] - terminal[0]
    direction_terminal[1] = point[1] - terminal[1]
    direction_terminal[2] = point[2] - terminal[2]
    upstream_length = sqrt(direction_upstream[0] ** 2.0 + direction_upstream[1] ** 2.0 + direction_upstream[2] ** 2.0)
    downstream_length = sqrt(direction_downstream[0] ** 2.0 + direction_downstream[1] ** 2.0 + direction_downstream[2] ** 2.0)
    terminal_length = sqrt(direction_terminal[0] ** 2.0 + direction_terminal[1] ** 2.0 + direction_terminal[2] ** 2.0)
    if upstream_length == 0.0:
        return HUGE_VAL #3.0*(M_PI*pow(data[0,21],2.0)*(data[0,20] + data[0,27]))
    elif downstream_length == 0.0:
        return HUGE_VAL #3.0*(M_PI*pow(data[0,21],2.0)*(data[0,20] + data[0,27]))
    elif terminal_length == 0.0:
        return HUGE_VAL #3.0*(M_PI*pow(data[0,21],2.0)*(data[0,20] + data[0,27]))
    r_terminal = ((8.0 * nu) / M_PI) * terminal_length
    r_terminal_sister = ((8.0 * nu) / M_PI) * downstream_length + (data[vessel, 25] - ((8.0 * nu) / M_PI) * data[vessel, 20])
    f_terminal = pow(1+pow(((data[vessel, 22] * r_terminal_sister) / (terminal_flow * r_terminal)), gamma/4.0),-1.0/gamma)
    f_terminal_sister = pow(1+pow(((data[vessel, 22]*r_terminal_sister) / (terminal_flow * r_terminal)), -gamma/4.0),-1.0/gamma)
    f_parent = 1.0
    r0 = ((8.0 * nu) / M_PI) * upstream_length + pow(((pow(f_terminal, 4.0))/r_terminal + (pow(f_terminal_sister, 4.0))/r_terminal_sister), -1.0)
    r_parent = r0
    l0 = pow(f_terminal, radius_exponent) * pow(terminal_length, length_exponent) + pow(f_terminal_sister, radius_exponent) * (pow(downstream_length, length_exponent) + data[vessel, 27])
    if vessel == 0:
        l0 = pow(upstream_length, length_exponent) + l0
        r0 = pow((((data[vessel, 22] + terminal_flow) * r0) / (root_pressure - terminal_pressure)), (1.0/4.0))
        cost = M_PI * pow(r0, radius_exponent) * l0
        return cost
    previous = vessel
    vessel = idxs[vessel, 2]
    if idxs[vessel, 0] == previous:
        alt = idxs[vessel, 1]
        if alt == -1:
            alt = -2
        position = 0
    else:
        alt = idxs[vessel, 0]
        if alt == -1:
            alt = -2
        position = 1
    if alt == -2:
        f_changed = 1.0
        f_stagnant = 0.0
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister * f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
        l0 = pow(f_changed, radius_exponent) * (pow(upstream_length, length_exponent) + l0)
    else:
        f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)),(gamma/4.0))), (-1.0/gamma))
        f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)),(-gamma/4.0))), (-1.0/gamma))
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + pow(((pow(f_changed, 4.0)) /r0 + (pow(f_stagnant, 4.0)) /data[alt, 25]), -1.0)
        l0 = pow(f_changed, radius_exponent) *(pow(upstream_length, length_exponent) + l0) + pow(f_stagnant, radius_exponent) * (pow(data[alt, 20], length_exponent) + data[alt, 27])
    if vessel == 0:
        l0 = pow(data[vessel, 20], length_exponent) + l0
        r0 = pow((((data[vessel, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
        cost = M_PI * pow(r0, radius_exponent) * l0
        return cost
    previous = vessel
    vessel = idxs[vessel, 2]
    while vessel >= 0:
        if idxs[vessel, 0] == previous:
            alt = idxs[vessel, 1]
            if alt == -1:
                alt = -2
            position = 0
        else:
            alt = idxs[vessel, 0]
            if alt == -1:
                alt = -2
            position = 1
        if alt == -2:
            f_changed = 1.0
            f_stagnant = 0.0
            r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
            l0 = pow(f_changed, radius_exponent) * (data[previous, 20] + l0)
            f_terminal = f_terminal * f_changed
            f_terminal_sister = f_terminal_sister * f_changed
            f_parent = f_parent * f_changed
        else:
            f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((terminal_flow+data[previous, 22])*r0)),
                                     (gamma/4.0))), (-1.0/gamma))
            f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt,25])/((terminal_flow+data[previous,22])*r0)),
                                      (-gamma/4.0))), (-1.0/gamma))
            f_terminal = f_terminal*f_changed
            f_terminal_sister = f_terminal_sister*f_changed
            f_parent = f_parent*f_changed
            r0 = ((8.0 * nu) / M_PI) * data[vessel,20] + pow(((pow(f_changed, 4.0) / r0) +
                                                              (pow(f_stagnant, 4.0) / data[alt, 25])), -1.0)
            l0 = pow(f_changed, radius_exponent) * (pow(data[previous, 20], length_exponent) + l0) + \
                 pow(f_stagnant, radius_exponent) *(pow(data[alt, 20], length_exponent) + data[alt, 27])
        previous = vessel
        vessel = idxs[vessel, 2]
    l0 = pow(data[0, 20], length_exponent) + l0
    r0 = pow((((data[0, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
    cost = M_PI * pow(r0, radius_exponent) * l0
    return cost

@cython.cpow(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def update_vessels_2(double[:] point, double[:,:] data, double[:] terminal, long[:,:] idxs,
                   int vessel, double gamma, double nu, double terminal_flow, double terminal_pressure,
                   double root_pressure, double radius_exponent, double length_exponent):
    cdef double upstream_length, downstream_length, terminal_length, r_terminal, r_terminal_sister, r_parent
    cdef double f_terminal, f_terminal_sister, r0, l0, f_changed, f_stagnant, f_parent, cost
    cdef vector[vector[double]] bifurcation_ratios
    cdef vector[double] tmp_bifurcation_ratios = [0.0, 0.0]
    cdef vector[int] main_idx
    cdef vector[int] alt_idx
    cdef vector[double] main_scale
    cdef vector[double] alt_scale
    cdef vector[double] reduced_resistance
    cdef vector[double] reduced_length
    cdef vector[double] flows
    cdef int previous, position, alt, alt_position
    cdef double[3] upstream_diff, downstream_diff, terminal_diff
    diff(point, data[vessel, 0:3], upstream_diff)
    diff(point, data[vessel, 3:6], downstream_diff)
    diff(point, terminal, terminal_diff)
    upstream_length = norm3(upstream_diff[0], upstream_diff[1], upstream_diff[2])
    downstream_length = norm3(downstream_diff[0], downstream_diff[1], downstream_diff[2])
    terminal_length = norm3(terminal_diff[0], terminal_diff[1], terminal_diff[2])
    r_terminal = ((8.0 * nu) / M_PI) * terminal_length
    r_terminal_sister = ((8.0 * nu) / M_PI) * downstream_length + (data[vessel, 25] -
                                                                   ((8.0 * nu) / M_PI) * data[vessel, 20])
    f_terminal = pow(1+pow(((data[vessel, 22] * r_terminal_sister) / (terminal_flow * r_terminal)), gamma/4.0),
                     -1.0/gamma)
    f_terminal_sister = pow(1+pow(((data[vessel, 22]*r_terminal_sister) / (terminal_flow * r_terminal)), -gamma/4.0),
                            -1.0/gamma)
    f_parent = 1.0
    r0 = ((8.0 * nu) / M_PI) * upstream_length + pow(((pow(f_terminal, 4.0))/r_terminal +
                                                     (pow(f_terminal_sister, 4.0))/r_terminal_sister), -1.0)
    r_parent = r0
    l0 = pow(f_terminal, radius_exponent) * pow(terminal_length, length_exponent) + \
         pow(f_terminal_sister, radius_exponent) * (pow(downstream_length, length_exponent) + data[vessel, 27])
    if vessel == 0:
        l0 = pow(upstream_length, length_exponent) + l0
        r0 = pow((((data[vessel, 22] + terminal_flow) * r0) / (root_pressure - terminal_pressure)), (1.0/4.0))
        return reduced_resistance, reduced_length, main_idx, main_scale, alt_idx, alt_scale, bifurcation_ratios, flows, r0, l0
    previous = vessel
    vessel = idxs[vessel, 2]
    if idxs[vessel, 0] == previous:
        alt = idxs[vessel, 1]
        if alt == -1:
            alt = -2
        position = 0
        alt_position = 1
    else:
        alt = idxs[vessel, 0]
        if alt == -1:
            alt = -2
        position = 1
        alt_position = 0
    if alt == -2:
        f_changed = 1.0
        f_stagnant = 0.0
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister * f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
        l0 = pow(f_changed, radius_exponent) * (pow(upstream_length, length_exponent) + l0)
    else:
        f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)),
                                 (gamma/4.0))), (-1.0/gamma))
        f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((data[previous, 22]+terminal_flow)*r0)),
                                  (-gamma/4.0))), (-1.0/gamma))
        f_terminal = f_terminal * f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_changed
        r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + pow(((pow(f_changed, 4.0)) /r0 +
                                                           (pow(f_stagnant, 4.0)) /data[alt, 25]), -1.0)
        l0 = pow(f_changed, radius_exponent) *(pow(upstream_length, length_exponent) + l0) + \
             pow(f_stagnant, radius_exponent) * (pow(data[alt, 20], length_exponent) + data[alt, 27])
    tmp_bifurcation_ratios[position] = f_changed
    tmp_bifurcation_ratios[alt_position] = f_stagnant
    if vessel == 0:
        main_idx.push_back(vessel)
        main_scale.push_back(1.0)
        alt_idx.push_back(alt)
        alt_scale.push_back(f_stagnant)
        reduced_resistance.push_back(r0)
        reduced_length.push_back(l0)
        tmp_bifurcation_ratios[position] = f_changed
        tmp_bifurcation_ratios[alt_position] = f_stagnant
        bifurcation_ratios.push_back(tmp_bifurcation_ratios)
        flows.push_back(data[vessel, 22] + terminal_flow)
        l0 = pow(data[vessel, 20], length_exponent) + l0
        r0 = pow((((data[vessel, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
        return reduced_resistance, reduced_length, main_idx, main_scale, alt_idx, alt_scale, bifurcation_ratios, flows, r0, l0
    else:
        main_idx.push_back(vessel)
        alt_idx.push_back(alt)
        main_scale.push_back(1.0)
        alt_scale.push_back(f_stagnant)
        reduced_resistance.push_back(r0)
        reduced_length.push_back(l0)
        tmp_bifurcation_ratios[position] = f_changed
        tmp_bifurcation_ratios[alt_position] = f_stagnant
        bifurcation_ratios.push_back(tmp_bifurcation_ratios)
        flows.push_back(data[vessel, 22] + terminal_flow)
    # All vessels past this point are above the triad
    previous = vessel
    vessel = idxs[vessel, 2]
    while vessel >= 0:
        if idxs[vessel, 0] == previous:
            alt = idxs[vessel, 1]
            if alt == -1:
                alt = -2
            position = 0
            alt_position = 1
        else:
            alt = idxs[vessel, 0]
            if alt == -1:
                alt = -2
            position = 1
            alt_position = 0
        if alt == -2:
            f_changed = 1.0
            f_stagnant = 0.0
            r0 = ((8.0 * nu) / M_PI) * data[vessel, 20] + r0
            l0 = pow(f_changed, radius_exponent) * (data[previous, 20] + l0)
            f_terminal = f_terminal * f_changed
            f_terminal_sister = f_terminal_sister * f_changed
            f_parent = f_parent * f_changed
        else:
            f_changed = pow((1.0+pow(((data[alt, 22]*data[alt, 25])/((terminal_flow+data[previous, 22])*r0)),
                                     (gamma/4.0))), (-1.0/gamma))
            f_stagnant = pow((1.0+pow(((data[alt, 22]*data[alt,25])/((terminal_flow+data[previous,22])*r0)),
                                      (-gamma/4.0))), (-1.0/gamma))
            f_terminal = f_terminal * f_changed
            f_terminal_sister = f_terminal_sister * f_changed
            f_parent = f_parent * f_changed
            r0 = ((8.0 * nu) / M_PI) * data[vessel,20] + pow(((pow(f_changed, 4.0) / r0) +
                                                              (pow(f_stagnant, 4.0) / data[alt, 25])), -1.0)
            l0 = pow(f_changed, radius_exponent) * (pow(data[previous, 20], length_exponent) + l0) + \
                 pow(f_stagnant, radius_exponent) *(pow(data[alt, 20], length_exponent) + data[alt, 27])
        for j in range(main_scale.size()):
            main_scale[j] = main_scale[j] * f_changed
            alt_scale[j] = alt_scale[j] * f_changed
        main_idx.push_back(vessel)
        alt_idx.push_back(alt)
        main_scale.push_back(1.0)
        alt_scale.push_back(f_stagnant)
        reduced_resistance.push_back(r0)
        reduced_length.push_back(l0)
        tmp_bifurcation_ratios[position] = f_changed
        tmp_bifurcation_ratios[alt_position] = f_stagnant
        bifurcation_ratios.push_back(tmp_bifurcation_ratios)
        flows.push_back(data[vessel, 22] + terminal_flow)
        previous = vessel
        vessel = idxs[vessel, 2]
    l0 = pow(data[0, 20], length_exponent) + l0
    r0 = pow((((data[0, 22]+terminal_flow)*r0)/(root_pressure-terminal_pressure)), (1.0/4.0))
    return reduced_resistance, reduced_length, main_idx, main_scale, alt_idx, alt_scale, bifurcation_ratios, flows, r0, l0


from xml.dom import minidom


class OutputBase:
    def __init__(self):
        self.output_type = None
        self.pressure = False
        self.velocity = False
        self.traction = False
        self.displacement = False
        self.wss = False
        self.temperature = False
        self.heat_flux = False
        self.alias = None

    def __str__(self):
        return str(vars(self))

    def __repr__(self):
        return str(vars(self))

    def __eq__(self, other):
        check = ["output_type", "pressure", "velocity", "traction", "displacement", "wss", "temperature", "heat_flux"]
        attributes = vars(self)
        other_attributes = vars(other)
        return all(other_attributes[key] == attributes[key] for key in check)

    def __ne__(self, other):
        return not self.__eq__(other)


class Output(OutputBase):
    def __init__(self):
        super().__init__()
        self.file = minidom.Document()

    def __str__(self):
        return self.toxml().toprettyxml()

    def __repr__(self):
        return self.toxml().toprettyxml()

    def set_type(self, output_type):
        if output_type not in ["Boundary_integral", "Spatial", "Volume_integral", "Alias"]:
            raise ValueError("Output type must be 'Boundary_integral', 'Spatial', or 'Volume_integral'.")
        self.output_type = output_type

    def toxml(self):
        output = self.file.createElement("Output")
        if not isinstance(self.output_type, type(None)):
            output.setAttribute("type", self.output_type)
        else:
            raise ValueError("Output type must be set.")

        if self.pressure:
            pressure = self.file.createElement("Pressure")
            pressure.appendChild(self.file.createTextNode("true"))
            output.appendChild(pressure)

        if self.velocity:
            velocity = self.file.createElement("Velocity")
            velocity.appendChild(self.file.createTextNode("true"))
            output.appendChild(velocity)

        if self.traction:
            traction = self.file.createElement("Traction")
            traction.appendChild(self.file.createTextNode("true"))
            output.appendChild(traction)

        if self.displacement:
            displacement = self.file.createElement("Displacement")
            displacement.appendChild(self.file.createTextNode("true"))
            output.appendChild(displacement)

        if self.wss:
            wss = self.file.createElement("WSS")
            wss.appendChild(self.file.createTextNode("true"))
            output.appendChild(wss)

        if self.temperature:
            temperature = self.file.createElement("Temperature")
            temperature.appendChild(self.file.createTextNode("true"))
            output.appendChild(temperature)

        if self.heat_flux:
            heat_flux = self.file.createElement("Heat_flux")
            heat_flux.appendChild(self.file.createTextNode("true"))
            output.appendChild(heat_flux)

        if self.output_type == 'Alias':
            if isinstance(self.alias, type(None)):
                raise ValueError("Alias must be set.")
            valid_outputs = ["Pressure", "Velocity", "Traction", "Displacement", "WSS", "Temperature", "Heat_flux"]
            for key, value in self.alias.items():
                if key not in valid_outputs:
                    raise ValueError("Alias key must be a valid output type. {} is not valid.".format(key))
                alias = self.file.createElement(key)
                alias.appendChild(self.file.createTextNode(value))
                output.appendChild(alias)
        return output



import tetgen
import pymeshfix
from svv.utils.remeshing import remesh


def triangulate(curve, verbose=False, **kwargs):
    """
    Triangulate a curve using VTK.

    Parameters
    ----------
    curve : Pyvista.PolyData PolyLine object
        The boundary curve within which the triangulation will
        be performed.
    verbose : bool
        A flag to indicate if mesh fixing should be verbose.
    kwargs : dict
        A dictionary of keyword arguments to be passed to VTK.

    Returns
    -------
    mesh : PyMesh mesh object
        A triangular mesh representing the triangulated region bounded by
        the curve.
    nodes : ndarray
        An array of node coordinates for the triangular mesh.
    vertices : ndarray
        An array of vertex indices for the triangular mesh.
    """
    mesh = curve.delaunay_2d(**kwargs)
    mesh = remesh.remesh_surface(mesh)
    nodes = mesh.points
    vertices = mesh.faces.reshape(-1, 4)[:, 1:]
    return mesh, nodes, vertices


def tetrahedralize(surface_mesh, verbose=False, **kwargs):
    """
    Tetrahedralize a surface mesh using TetGen.

    Parameters
    ----------
    surface_mesh : PyMesh mesh object
        The surface mesh to tetrahedralize.
    verbose : bool
        A flag to indicate if mesh fixing should be verbose.
    kwargs : dict
        A dictionary of keyword arguments to be passed to TetGen.

    Returns
    -------
    mesh : PyMesh mesh object
        An unstructured grid mesh representing the tetrahedralized
        volume enclosed by the surface mesh manifold.
    """
    mesh = pymeshfix.MeshFix(surface_mesh)
    mesh.repair(verbose=verbose)
    tet = tetgen.TetGen(mesh.mesh)
    nodes, vertices = tet.tetrahedralize(**kwargs)
    mesh = tet.grid
    return mesh, nodes, vertices



# cython: language_level=3
# distutils: language=c++
import numpy as np
cimport numpy as cnp
import cython
from libc.math cimport fabs

cnp.import_array()

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef dot(double[:] a, double[:] b):
    cdef Py_ssize_t i = a.shape[0]
    cdef double res = 0.0
    for ii in range(i):
        res += a[ii] * b[ii]
    return res

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef dot_inplace(double[:] a, double[:] b, double res):
    cdef Py_ssize_t i = a.shape[0]
    res = 0.0
    for ii in range(i):
        res += a[ii] * b[ii]

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef cross(double[:] a, double[:] b, double[:] res):
    res[0] = a[1] * b[2] - a[2] * b[1]
    res[1] = a[2] * b[0] - a[0] * b[2]
    res[2] = a[0] * b[1] - a[1] * b[0]

#@cython.boundscheck(False)
#@cython.wraparound(False)
#@cython.nonecheck(False)
#cdef int separating_axis(double[:] position, double[:] plane,
#                     double[:] u1, double[:] v1, double[:] w1,
#                     double[:] u2, double[:] v2, double[:] w2,
#                     double u1s, double v1s, double w1s,
#                     double u2s, double v2s, double w2s):
#    if fabs(dot(position, plane)) > (fabs(u1s*dot(u1, plane)) +
#                                        fabs(v1s*dot(v1, plane)) +
#                                        fabs(w1s*dot(w1, plane)) +
#                                        fabs(u2s*dot(u2, plane)) +
#                                        fabs(v2s*dot(v2, plane)) +
#                                        fabs(w2s*dot(w2, plane))):
#        return 1
#    else:
#        return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef int separating_axis(double[3] position, double[3] plane,
                         double[3] u1, double[3] v1, double[3] w1,
                         double[3] u2, double[3] v2, double[3] w2,
                         double u1s, double v1s, double w1s,
                         double u2s, double v2s, double w2s) nogil:
    cdef double lhs, rhs
    cdef double pos_dot_plane
    cdef double u1_dot_plane, v1_dot_plane, w1_dot_plane
    cdef double u2_dot_plane, v2_dot_plane, w2_dot_plane

    # Compute dot products directly
    pos_dot_plane = position[0]*plane[0] + position[1]*plane[1] + position[2]*plane[2]
    u1_dot_plane = u1[0]*plane[0] + u1[1]*plane[1] + u1[2]*plane[2]
    v1_dot_plane = v1[0]*plane[0] + v1[1]*plane[1] + v1[2]*plane[2]
    w1_dot_plane = w1[0]*plane[0] + w1[1]*plane[1] + w1[2]*plane[2]
    u2_dot_plane = u2[0]*plane[0] + u2[1]*plane[1] + u2[2]*plane[2]
    v2_dot_plane = v2[0]*plane[0] + v2[1]*plane[1] + v2[2]*plane[2]
    w2_dot_plane = w2[0]*plane[0] + w2[1]*plane[1] + w2[2]*plane[2]

    # Compute the left-hand side (lhs) and right-hand side (rhs) of the inequality
    lhs = fabs(pos_dot_plane)
    rhs = (fabs(u1s * u1_dot_plane) +
           fabs(v1s * v1_dot_plane) +
           fabs(w1s * w1_dot_plane) +
           fabs(u2s * u2_dot_plane) +
           fabs(v2s * v2_dot_plane) +
           fabs(w2s * w2_dot_plane))

    # Return the result based on the comparison
    if lhs > rhs:
        return 1
    else:
        return 0

"""
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
def obb_any(double[:,:] data, double[:,:] vessels):
    
    This Oriented Bounding Box (OBB) collision detection algorithm is based on the Separating Axis Theorem (SAT).
    It is used to determine if any of the OBBs in the data array are colliding with any of the OBBs in the vessels array.
    This algorithm returns True at the first instance of a collision and returns False if, and only if, no collision
    exists among the OBBs in the data and vessel arrays.

    Parameters
    ----------
    data : numpy.ndarray
        A 2D numpy array of shape (n, j) where n is the number of OBBs in the data array. The columns of the data
        array represent the 3D position, orientation, and half-widths of the OBBs in the data array.
    vessels : numpy.ndarray
        A 2D numpy array of shape (m, j) where m is the number of OBBs in the vessels array. The columns of the
        vessels array represent the 3D position, orientation, and half-widths of the OBBs in the vessels array.

    Returns
    -------
    has_collision : bool
        A boolean value indicating if any of the OBBs in the data array are colliding with any of the OBBs in the
        vessels array.
    
    cdef Py_ssize_t j = data.shape[0]
    cdef Py_ssize_t i = vessels.shape[0]
    cdef int k = 3
    cdef int kk
    #cdef cnp.ndarray[cnp.float64_t, ndim=1] center_1 = np.zeros(k, dtype=np.float64)
    #cdef cnp.ndarray[cnp.float64_t, ndim=1] center_2 = np.zeros(k, dtype=np.float64)
    #cdef cnp.ndarray[cnp.float64_t, ndim=1] position = np.zeros(k, dtype=np.float64)
    #cdef cnp.ndarray[cnp.float64_t, ndim=1] plane = np.zeros(k, dtype=np.float64)
    #cdef double[:] center_1_mv = center_1
    #cdef double[:] center_2_mv = center_2
    #cdef double[:] position_mv = position
    #cdef double[:] plane_mv = plane
    cdef double[3] center_1_mv = [0.0, 0.0, 0.0]
    cdef double[3] center_2_mv = [0.0, 0.0, 0.0]
    cdef double[3] position_mv = [0.0, 0.0, 0.0]
    cdef double[3] plane_mv = [0.0, 0.0, 0.0]
    for ii in range(i):
        for kk in range(k):
            center_1_mv[kk] = (vessels[ii, kk] + vessels[ii, kk + k]) / 2.0
        for jj in range(j):
            for kk in range(k):
                center_2_mv[kk] = (data[jj, kk] + data[jj, kk + k]) / 2.0
                position_mv[kk] = center_2_mv[kk] - center_1_mv[kk]
            if separating_axis(position_mv, vessels[ii,6:9], vessels[ii, 6:9], vessels[ii,9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            if separating_axis(position_mv, vessels[ii, 9:12], vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # W1 axis
            if separating_axis(position_mv, vessels[ii, 12:15], vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # U2 axis
            if separating_axis(position_mv, data[jj, 6:9], vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # V2 axis
            if separating_axis(position_mv, data[jj, 9:12], vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # W2 axis
            if separating_axis(position_mv, data[jj, 12:15], vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # Mixed plane Separation Tests
            # U1 X U2
            cross(vessels[ii, 6:9], data[jj, 6:9], plane_mv)
            if separating_axis(position_mv, plane_mv, vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # U1 X V2
            cross(vessels[ii, 6:9], data[jj, 9:12], plane_mv)
            if separating_axis(position_mv, plane_mv, vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # U1 X W2
            cross(vessels[ii, 6:9], data[jj, 12:15], plane_mv)
            if separating_axis(position_mv, plane_mv, vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # V1 X U2
            cross(vessels[ii, 9:12], data[jj, 6:9], plane_mv)
            if separating_axis(position_mv, plane_mv, vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # V1 X V2
            cross(vessels[ii, 9:12], data[jj, 9:12], plane_mv)
            if separating_axis(position_mv, plane_mv, vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # V1 X W2
            cross(vessels[ii, 9:12], data[jj, 12:15], plane_mv)
            if separating_axis(position_mv, plane_mv, vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # W1 X U2
            cross(vessels[ii, 12:15], data[jj, 6:9], plane_mv)
            if separating_axis(position_mv, plane_mv, vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # W1 X V2
            cross(vessels[ii, 12:15], data[jj, 9:12], plane_mv)
            if separating_axis(position_mv, plane_mv, vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            # W1 X W2
            cross(vessels[ii, 12:15], data[jj, 12:15], plane_mv)
            if separating_axis(position_mv, plane_mv, vessels[ii, 6:9], vessels[ii, 9:12], vessels[ii, 12:15],
                               data[jj, 6:9], data[jj, 9:12], data[jj, 12:15], vessels[ii, 21], vessels[ii, 21],
                               vessels[ii, 20] / 2, data[jj, 21], data[jj, 21], data[jj, 20] / 2):
                continue
            else:
                return True
    return False
"""

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
def obb_any(double[:,:] data, double[:,:] vessels):
    """
    Oriented Bounding Box (OBB) collision detection using the Separating Axis Theorem (SAT).
    Returns True if any collision is detected between the OBBs in 'data' and 'vessels'.
    """
    cdef Py_ssize_t num_data = data.shape[0]
    cdef Py_ssize_t num_vessels = vessels.shape[0]
    cdef Py_ssize_t i_data, i_vessel
    cdef double[3] center1, center2, position
    cdef double[3] u1, v1, w1, u2, v2, w2
    cdef double u1s, v1s, w1s, u2s, v2s, w2s
    cdef double[3] cross_prod
    cdef int collision

    with nogil:
        for i_vessel in range(num_vessels):
            # Extract vessel OBB center
            center1[0] = (vessels[i_vessel, 0] + vessels[i_vessel, 3]) * 0.5
            center1[1] = (vessels[i_vessel, 1] + vessels[i_vessel, 4]) * 0.5
            center1[2] = (vessels[i_vessel, 2] + vessels[i_vessel, 5]) * 0.5

            # Extract vessel OBB axes
            u1[0] = vessels[i_vessel, 6]
            u1[1] = vessels[i_vessel, 7]
            u1[2] = vessels[i_vessel, 8]

            v1[0] = vessels[i_vessel, 9]
            v1[1] = vessels[i_vessel,10]
            v1[2] = vessels[i_vessel,11]

            w1[0] = vessels[i_vessel,12]
            w1[1] = vessels[i_vessel,13]
            w1[2] = vessels[i_vessel,14]

            # Vessel OBB half-sizes
            u1s = vessels[i_vessel, 21]
            v1s = vessels[i_vessel, 21]
            w1s = vessels[i_vessel, 20] * 0.5

            for i_data in range(num_data):
                # Extract data OBB center
                center2[0] = (data[i_data, 0] + data[i_data, 3]) * 0.5
                center2[1] = (data[i_data, 1] + data[i_data, 4]) * 0.5
                center2[2] = (data[i_data, 2] + data[i_data, 5]) * 0.5

                # Compute vector between centers
                position[0] = center2[0] - center1[0]
                position[1] = center2[1] - center1[1]
                position[2] = center2[2] - center1[2]

                # Extract data OBB axes
                u2[0] = data[i_data, 6]
                u2[1] = data[i_data, 7]
                u2[2] = data[i_data, 8]

                v2[0] = data[i_data, 9]
                v2[1] = data[i_data,10]
                v2[2] = data[i_data,11]

                w2[0] = data[i_data,12]
                w2[1] = data[i_data,13]
                w2[2] = data[i_data,14]

                # Data OBB half-sizes
                u2s = data[i_data, 21]
                v2s = data[i_data, 21]
                w2s = data[i_data, 20] * 0.5

                # Begin testing axes for separation
                # Test vessel axes
                if separating_axis(position, u1, u1, v1, w1, u2, v2, w2,
                                   u1s, v1s, w1s, u2s, v2s, w2s):
                    continue  # Separating axis found

                if separating_axis(position, v1, u1, v1, w1, u2, v2, w2,
                                   u1s, v1s, w1s, u2s, v2s, w2s):
                    continue

                if separating_axis(position, w1, u1, v1, w1, u2, v2, w2,
                                   u1s, v1s, w1s, u2s, v2s, w2s):
                    continue

                # Test data axes
                if separating_axis(position, u2, u1, v1, w1, u2, v2, w2,
                                   u1s, v1s, w1s, u2s, v2s, w2s):
                    continue

                if separating_axis(position, v2, u1, v1, w1, u2, v2, w2,
                                   u1s, v1s, w1s, u2s, v2s, w2s):
                    continue

                if separating_axis(position, w2, u1, v1, w1, u2, v2, w2,
                                   u1s, v1s, w1s, u2s, v2s, w2s):
                    continue

                # Test cross products of axes
                # U1 x U2
                cross_prod[0] = u1[1] * u2[2] - u1[2] * u2[1]
                cross_prod[1] = u1[2] * u2[0] - u1[0] * u2[2]
                cross_prod[2] = u1[0] * u2[1] - u1[1] * u2[0]

                if not (cross_prod[0] == 0 and cross_prod[1] == 0 and cross_prod[2] == 0):
                    if separating_axis(position, cross_prod, u1, v1, w1, u2, v2, w2,
                                       u1s, v1s, w1s, u2s, v2s, w2s):
                        continue

                # U1 x V2
                cross_prod[0] = u1[1] * v2[2] - u1[2] * v2[1]
                cross_prod[1] = u1[2] * v2[0] - u1[0] * v2[2]
                cross_prod[2] = u1[0] * v2[1] - u1[1] * v2[0]

                if not (cross_prod[0] == 0 and cross_prod[1] == 0 and cross_prod[2] == 0):
                    if separating_axis(position, cross_prod, u1, v1, w1, u2, v2, w2,
                                       u1s, v1s, w1s, u2s, v2s, w2s):
                        continue

                # U1 x W2
                cross_prod[0] = u1[1] * w2[2] - u1[2] * w2[1]
                cross_prod[1] = u1[2] * w2[0] - u1[0] * w2[2]
                cross_prod[2] = u1[0] * w2[1] - u1[1] * w2[0]

                if not (cross_prod[0] == 0 and cross_prod[1] == 0 and cross_prod[2] == 0):
                    if separating_axis(position, cross_prod, u1, v1, w1, u2, v2, w2,
                                       u1s, v1s, w1s, u2s, v2s, w2s):
                        continue

                # V1 x U2
                cross_prod[0] = v1[1] * u2[2] - v1[2] * u2[1]
                cross_prod[1] = v1[2] * u2[0] - v1[0] * u2[2]
                cross_prod[2] = v1[0] * u2[1] - v1[1] * u2[0]

                if not (cross_prod[0] == 0 and cross_prod[1] == 0 and cross_prod[2] == 0):
                    if separating_axis(position, cross_prod, u1, v1, w1, u2, v2, w2,
                                       u1s, v1s, w1s, u2s, v2s, w2s):
                        continue

                # V1 x V2
                cross_prod[0] = v1[1] * v2[2] - v1[2] * v2[1]
                cross_prod[1] = v1[2] * v2[0] - v1[0] * v2[2]
                cross_prod[2] = v1[0] * v2[1] - v1[1] * v2[0]

                if not (cross_prod[0] == 0 and cross_prod[1] == 0 and cross_prod[2] == 0):
                    if separating_axis(position, cross_prod, u1, v1, w1, u2, v2, w2,
                                       u1s, v1s, w1s, u2s, v2s, w2s):
                        continue

                # V1 x W2
                cross_prod[0] = v1[1] * w2[2] - v1[2] * w2[1]
                cross_prod[1] = v1[2] * w2[0] - v1[0] * w2[2]
                cross_prod[2] = v1[0] * w2[1] - v1[1] * w2[0]

                if not (cross_prod[0] == 0 and cross_prod[1] == 0 and cross_prod[2] == 0):
                    if separating_axis(position, cross_prod, u1, v1, w1, u2, v2, w2,
                                       u1s, v1s, w1s, u2s, v2s, w2s):
                        continue

                # W1 x U2
                cross_prod[0] = w1[1] * u2[2] - w1[2] * u2[1]
                cross_prod[1] = w1[2] * u2[0] - w1[0] * u2[2]
                cross_prod[2] = w1[0] * u2[1] - w1[1] * u2[0]

                if not (cross_prod[0] == 0 and cross_prod[1] == 0 and cross_prod[2] == 0):
                    if separating_axis(position, cross_prod, u1, v1, w1, u2, v2, w2,
                                       u1s, v1s, w1s, u2s, v2s, w2s):
                        continue

                # W1 x V2
                cross_prod[0] = w1[1] * v2[2] - w1[2] * v2[1]
                cross_prod[1] = w1[2] * v2[0] - w1[0] * v2[2]
                cross_prod[2] = w1[0] * v2[1] - w1[1] * v2[0]

                if not (cross_prod[0] == 0 and cross_prod[1] == 0 and cross_prod[2] == 0):
                    if separating_axis(position, cross_prod, u1, v1, w1, u2, v2, w2,
                                       u1s, v1s, w1s, u2s, v2s, w2s):
                        continue

                # W1 x W2
                cross_prod[0] = w1[1] * w2[2] - w1[2] * w2[1]
                cross_prod[1] = w1[2] * w2[0] - w1[0] * w2[2]
                cross_prod[2] = w1[0] * w2[1] - w1[1] * w2[0]

                if not (cross_prod[0] == 0 and cross_prod[1] == 0 and cross_prod[2] == 0):
                    if separating_axis(position, cross_prod, u1, v1, w1, u2, v2, w2,
                                       u1s, v1s, w1s, u2s, v2s, w2s):
                        continue
                with gil:
                    # No separating axis found; collision detected
                    return True
    # No collisions detected
    return False



import numpy
import pyvista as pv
from copy import deepcopy
from svv.forest.connect.base_connection import BaseConnection


class VesselConnection:
    def __init__(self, forest, network_id, tree_0, tree_1, idx, jdx,
                 ctrl_function=None, clamp_first=True, clamp_second=True,
                 point_0=None, point_1=None, curve_type="Bezier"):
        self.forest = forest
        self.tree_0 = tree_0
        self.tree_1 = tree_1
        self.idx = idx
        self.jdx = jdx
        vessel_0 = forest.networks[network_id][tree_0].data[idx, :]
        vessel_1 = forest.networks[network_id][tree_1].data[jdx, :]
        self.proximal_0 = vessel_0[0:3]
        self.distal_0 = vessel_0[3:6]
        self.proximal_1 = vessel_1[0:3]
        self.distal_1 = vessel_1[3:6]
        self.radius_0 = vessel_0[21]
        self.radius_1 = vessel_1[21]
        collision_vessels = []
        min_distance = max(self.radius_0, self.radius_1)*0.5
        #min_distance = (max(numpy.max(forest.networks[network_id][tree_0].data[:,21]),
        #                   numpy.max(forest.networks[network_id][tree_1].data[:,21])) + max(self.radius_0,self.radius_1))
        conn = BaseConnection(vessel_0[0:3], vessel_0[3:6], vessel_1[0:3], vessel_1[3:6], vessel_0[21], vessel_1[21],
                              domain=forest.domain, ctrlpt_function=ctrl_function, clamp_first=clamp_first,
                              clamp_second=clamp_second, point_0=point_0, point_1=point_1, min_distance=min_distance,
                              curve_type=curve_type)
        tree_0_idx = numpy.arange(forest.networks[network_id][tree_0].data.shape[0], dtype=int).tolist()
        if not numpy.isnan(vessel_0[17]):
            parent = int(vessel_0[17])
            daughter_0 = int(forest.networks[network_id][tree_0].data[parent, 15])
            daughter_1 = int(forest.networks[network_id][tree_0].data[parent, 16])
            #tree_0_idx.remove(parent)
            #if daughter_0 == idx:
            #    tree_0_idx.remove(daughter_0)
            #else:
            #    tree_0_idx.remove(daughter_1)
            tree_0_idx.remove(idx)
        if not numpy.isnan(vessel_0[15]):
            daughter_2 = int(vessel_1[15])
            tree_0_idx.remove(daughter_2)
        if not numpy.isnan(vessel_0[16]):
            daughter_3 = int(vessel_1[16])
            tree_0_idx.remove(daughter_3)
        tree_0_idx = numpy.array(tree_0_idx).astype(int)
        tmp = numpy.zeros((tree_0_idx.shape[0], 7))
        tmp[:, 0:3] = forest.networks[network_id][tree_0].data[tree_0_idx, 0:3]
        tmp[:, 3:6] = forest.networks[network_id][tree_0].data[tree_0_idx, 3:6]
        tmp[:, 6] = forest.networks[network_id][tree_0].data[tree_0_idx, 21]
        collision_vessels.append(deepcopy(tmp))
        tree_1_idx = numpy.arange(forest.networks[network_id][tree_1].data.shape[0], dtype=int).tolist()
        if not numpy.isnan(vessel_1[17]):
            parent = int(vessel_1[17])
            daughter_0 = int(forest.networks[network_id][tree_1].data[parent, 15])
            daughter_1 = int(forest.networks[network_id][tree_1].data[parent, 16])
            #tree_1_idx.remove(parent)
            #if daughter_0 == jdx:
            #    tree_1_idx.remove(daughter_0)
            #else:
            #    tree_1_idx.remove(daughter_1)
            #tree_1_idx.remove(daughter_0)
            #tree_1_idx.remove(daughter_1)
            tree_1_idx.remove(jdx)
        if not numpy.isnan(vessel_1[15]):
            daughter_2 = int(vessel_1[15])
            tree_1_idx.remove(daughter_2)
        if not numpy.isnan(vessel_1[16]):
            daughter_3 = int(vessel_1[16])
            tree_1_idx.remove(daughter_3)
        tree_1_idx = numpy.array(tree_1_idx).astype(int)
        tmp = numpy.zeros((tree_1_idx.shape[0], 7))
        tmp[:, 0:3] = forest.networks[network_id][tree_1].data[tree_1_idx, 0:3]
        tmp[:, 3:6] = forest.networks[network_id][tree_1].data[tree_1_idx, 3:6]
        tmp[:, 6] = forest.networks[network_id][tree_1].data[tree_1_idx, 21]
        collision_vessels.append(deepcopy(tmp))
        for i in range(forest.n_networks):
            for j in range(forest.n_trees_per_network[i]):
                if i == network_id and (j == tree_0 or j == tree_1):
                    continue
                tmp = numpy.zeros((forest.networks[i][j].data.shape[0], 7))
                tmp[:, 0:3] = forest.networks[i][j].data[:, 0:3]
                tmp[:, 3:6] = forest.networks[i][j].data[:, 3:6]
                tmp[:, 6] = forest.networks[i][j].data[:, 21]
                collision_vessels.append(deepcopy(tmp))
        collision_vessels = numpy.vstack(collision_vessels)
        if collision_vessels.shape[0] > 0:
            conn.set_collision_vessels(collision_vessels)
        conn.set_physical_clearance(self.forest.networks[network_id][tree_0].domain_clearance)
        bounds = numpy.zeros((3, 2))
        bounds[:, 0] = numpy.min(forest.domain.points, axis=0).T
        bounds[:, 1] = numpy.max(forest.domain.points, axis=0).T
        conn.clamp_first = clamp_first
        conn.clamp_second = clamp_second
        self.connection = conn
        self.curve = None
        self.result = None
        self.vessels_1 = None
        self.vessels_2 = None
        self.vessel_1_meshes = None
        self.vessel_2_meshes = None
        self.seperate = None
        self.plotting_vessels = None

    def solve(self, *args, **kwargs):
        result, curve, objective, constraints = self.connection.solve(*args, **kwargs)
        self.result = result
        self.curve = curve
        self.objective = objective
        self.constraints = constraints

    def build_vessels(self, num, seperate=True, build_meshes=True):
        t = numpy.linspace(0, 1, num)
        pts = self.curve.evaluate(t)
        if seperate:
            sep = (pts.shape[0]-1)//2
            vessels_1 = numpy.zeros((sep,7))
            vessels_2 = numpy.zeros((pts.shape[0]-1-sep,7))
            vessels_1_pts = pts[:sep+1,:]
            vessels_2_pts = numpy.flip(pts[sep:,:],axis=0)
            vessels_1[:, 0:3] = vessels_1_pts[:-1,:]
            vessels_1[:, 3:6] = vessels_1_pts[1:,:]
            vessels_2[:, 0:3] = vessels_2_pts[:-1,:]
            vessels_2[:, 3:6] = vessels_2_pts[1:,:]
            vessels_1[:, 6] = self.radius_0
            vessels_2[:, 6] = self.radius_1
        else:
            vessels_1 = numpy.zeros((pts.shape[0]-1,7))
            vessels_1[:,0:3] = pts[:-1,:]
            vessels_1[:,3:6] = pts[1:,:]
            vessels_1[:,6] = self.radius_0
            pts_reverse = numpy.flip(pts,axis=0)
            vessels_2 = numpy.zeros((pts.shape[0]-1,7))
            vessels_2[:,0:3] = pts_reverse[:-1,:]
            vessels_2[:,3:6] = pts_reverse[1:,:]
            vessels_2[:,6] = self.radius_1
        self.vessels_1 = vessels_1
        self.vessels_2 = vessels_2
        self.seperate = seperate
        self.vessel_1_meshes = []
        if build_meshes:
            for i in range(vessels_1.shape[0]):
                center = (self.vessels_1[i, 0:3] + self.vessels_1[i, 3:6]) / 2
                direction = self.vessels_1[i, 3:6] - self.vessels_1[i, 0:3]
                length = numpy.linalg.norm(direction)
                direction = direction / length
                radius = self.vessels_1[i, 6]
                cylinder = pv.Cylinder(center=center, direction=direction, radius=radius, height=length)
                self.vessel_1_meshes.append(cylinder)
            self.vessel_2_meshes = []
            for i in range(self.vessels_2.shape[0]):
                center = (self.vessels_2[i, 0:3] + self.vessels_2[i, 3:6]) / 2
                direction = self.vessels_2[i, 3:6] - self.vessels_2[i, 0:3]
                length = numpy.linalg.norm(direction)
                direction = direction / length
                radius = self.vessels_2[i, 6]
                cylinder = pv.Cylinder(center=center, direction=direction, radius=radius, height=length)
                self.vessel_2_meshes.append(cylinder)
        return vessels_1, vessels_2

    def show(self):
        plotter = pv.Plotter()
        if self.seperate:
            for i in range(self.vessels_1.shape[0]):
                center = (self.vessels_1[i,0:3]+self.vessels_1[i,3:6])/2
                direction = self.vessels_1[i,3:6] - self.vessels_1[i,0:3]
                length = numpy.linalg.norm(direction)
                direction = direction/length
                radius = self.vessels_1[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='red')
            for i in range(self.vessels_2.shape[0]):
                center = (self.vessels_2[i,0:3]+self.vessels_2[i,3:6])/2
                direction = self.vessels_2[i,3:6] - self.vessels_2[i,0:3]
                length = numpy.linalg.norm(direction)
                direction = direction/length
                radius = self.vessels_2[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='blue')
        else:
            for i in range(self.vessels_1.shape[0]):
                center = (self.vessels_1[i,0:3]+self.vessels_1[i,3:6])/2
                direction = self.vessels_1[i,3:6] - self.vessels_1[i,0:3]
                length = numpy.linalg.norm(direction)
                direction = direction/length
                radius = self.vessels_1[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='red')
        if self.plotting_vessels is not None:
            for i in range(self.plotting_vessels.shape[0]):
                center = (self.plotting_vessels[i,0:3]+self.plotting_vessels[i,3:6])/2
                direction = self.plotting_vessels[i,3:6] - self.plotting_vessels[i,0:3]
                length = numpy.linalg.norm(direction)
                direction = direction/length
                radius = self.plotting_vessels[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='black')
        if numpy.all(self.proximal_0 != self.distal_0):
            center = (self.proximal_0+self.distal_0)/2
            direction = (self.distal_0-self.proximal_0)
            length = numpy.linalg.norm(direction)
            direction = direction/length
            radius = self.radius_0
            cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
            plotter.add_mesh(cylinder,color='yellow')
        if numpy.all(self.proximal_1 != self.distal_1):
            center = (self.proximal_1+self.distal_1)/2
            direction = (self.distal_1-self.proximal_1)
            length = numpy.linalg.norm(direction)
            direction = direction/length
            radius = self.radius_1
            cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
            plotter.add_mesh(cylinder,color='yellow')
        return plotter



import numpy as np
from .m_matrix import m_matrix
from .n_matrix import n_matrix


def a_matrix(points, rbf_degree=3):
    """
    Construct the full interpolation matrix for Duchon spline interpolation.

    This function builds the full interpolation matrix `A` used in Duchon spline interpolation
    over a set of points defining a (D-1)-dimensional manifold embedded in D-dimensional space.
    The matrix `A` is assembled from the sub-matrices `M` and `N`, which are constructed based
    on the input points and the specified radial basis function (RBF) degree.


    Parameters
    ----------
    points : ndarray of shape (n_points, dim)
        Coordinates of the points defining the manifold. Each row corresponds to a point
        in D-dimensional space.

    rbf_degree : int or float, optional
        Degree of the Duchon interpolant function. Controls the smoothness of the interpolation.
        Default is tri-harmonic kernel rbf_degree=3.

    Returns
    -------
    A : ndarray of shape ((n_points + 1) * (dim + 1), (n_points + 1) * (dim + 1))
        The full interpolation matrix assembled from sub-matrices `M` and `N`.
        This matrix is used to compute the interpolation coefficients for the Duchon spline.

    Notes
    -----
    Duchon splines generalize thin-plate splines to higher dimensions, allowing interpolation
    on manifolds of arbitrary dimension [1]_. The construction of the interpolation matrix involves
    creating two sub-matrices:

    - **M**: A block matrix representing the interaction between points via the RBF.
    - **N**: A matrix incorporating polynomial terms to ensure the interpolation passes through
      the given points.

    The full matrix `A` is assembled as:

    .. math::

        A = \\begin{bmatrix}
                M & N \\\\
                N^T & 0
            \\end{bmatrix}

    where the zero block is of appropriate size to complete the matrix.

    References
    ----------
    .. [1] Duchon, J. (1977). Splines minimizing rotation-invariant semi-norms in Sobolev spaces.
           In *Constructive Theory of Functions of Several Variables* (pp. 85-100). Springer, Berlin, Heidelberg.
           `Springer Link <https://doi.org/10.1007/BFb0086566>`_

    Examples
    --------
    **Example 1: Using 2D Points**

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.core.a_matrix import a_matrix

        # Define a set of points in 2D space
        points = np.array([
            [0.0, 0.0],
            [1.0, 0.0],
            [0.0, 1.0],
            [1.0, 1.0]
        ])

        # Compute the interpolation matrix with the default RBF degree
        A = a_matrix(points)

        print("Shape of the interpolation matrix:", A.shape)
        # Output: Shape of the interpolation matrix: (20, 20)

    **Example 2: Using 3D Points**

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.core.a_matrix import a_matrix

        # Define a set of 3D points
        points = np.array([
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 1.0, 1.0]
        ])

        # Compute the interpolation matrix for 3D points
        A = a_matrix(points)

        print("Shape of the interpolation matrix:", A.shape)
        # Output: Shape of the interpolation matrix: (24, 24)

    **Example 3: Changing the RBF Degree**

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.core.a_matrix import a_matrix

        # Define a set of 2D points
        points = np.array([
            [0.0, 0.0],
            [0.5, 0.5],
            [1.0, 0.0],
            [0.0, 1.0],
            [1.0, 1.0]
        ])

        # Compute the interpolation matrix with a higher RBF degree
        A_high_degree = a_matrix(points, rbf_degree=5)

        # Compute the interpolation matrix with a lower RBF degree
        A_low_degree = a_matrix(points, rbf_degree=2)

        print("Shape with higher degree (5):", A_high_degree.shape)
        print("Shape with lower degree (2):", A_low_degree.shape)
        # Output:
        # Shape with higher degree (5): (24, 24)
        # Shape with lower degree (2): (24, 24)

    **Example 4: Practical Interpolation Task**

    Suppose you have a function defined on a set of 2D points, and you want to interpolate this function over the entire plane.

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.core.a_matrix import a_matrix
        from scipy.linalg import solve

        # Define a set of 2D points
        points = np.array([
            [0.0, 0.0],
            [1.0, 0.0],
            [0.0, 1.0],
            [1.0, 1.0],
            [0.5, 0.5]
        ])

        # Values of the function at the given points
        values = np.array([0, 1, 1, 0, 0.5])

        # Build the interpolation matrix
        A = a_matrix(points)

        # Right-hand side vector
        b = np.zeros(A.shape[0])
        b[:len(values)*(points.shape[1]+1):points.shape[1]+1] = values

        # Solve for the interpolation coefficients
        coeffs = solve(A, b)

        print("Interpolation coefficients:")
        print(coeffs)

    See Also
    --------
    :func:`svtoolkit.domain.core.h_matrix.h_matrix` : Function to compute the H matrix for bending energy minimization.

    """
    n = points.shape[0]
    d = points.shape[1]
    m_ = m_matrix(points, rbf_degree=rbf_degree)
    n_ = n_matrix(points)
    a_ = np.zeros(((n + 1) * (d + 1), (n + 1) * (d + 1)))
    a_[:n * (d + 1), :n * (d + 1)] = m_
    a_[:n * (d + 1), n * (d + 1):] = n_
    a_[n * (d + 1):, :n * (d + 1)] = n_.T
    return a_



import numpy as np
from scipy import optimize
from functools import partial
from ..kernel.coordinate_system import sph2cart


class DomainSolution:
    def __init__(self):
        self.x = None
        self.fun = None


class Solver:
    """
    This class defines the solver for the implicit domain object.
    """

    def __init__(self, kernel):
        self.kernel = kernel
        self.algorithm = None
        self.solution = None
        self.solutions = None

    def set_solver(self, method='L-BFGS-B', verbose=False):
        """
        This function sets the solver method to be used for the optimization problem.
        """
        scipy_solvers = ['Nelder-Mead', 'Powell', 'CG', 'BFGS', 'Newton-CG',
                         'L-BFGS-B', 'TNC', 'COBYLA', 'SLSQP', 'trust-constr',
                         'dogleg', 'trust-ncg', 'trust-exact', 'trust-krylov']
        if method in scipy_solvers:
            if verbose:
                if method == 'trust-constr':
                    options = {'disp': True}
                elif method == 'L-BFGS-B':
                    options = {'disp': 99, 'maxls': 40, 'ftol': 1e-9}
                elif method in ['Newton-CG', 'CG', 'BFGS', 'Nelder-Mead',
                                'Powell', 'TNC', 'COBYLA', 'SLSQP']:
                    options = {'disp': True}
                else:
                    print('No verbosity allowed for this method.')
                    options = {}
            else:
                options = {}
            algorithm = partial(optimize.minimize, method=method, tol=1e-09, options=options)
        else:
            print('Not an available solver method.')
            print('See scipy methods: {}'.format(scipy_solvers))
            algorithm = None
        self.algorithm = algorithm

    def solve(self, skip=False):
        if self.algorithm is None:
            print('Solver not set.')
            return None
        if skip:
            solution = DomainSolution()
            solution.fun = self.kernel.__cost__(self.kernel.x0[0])
            solution.x = self.kernel.x0[0]
            self.solution = solution
            self.solutions = [solution]
            return solution
        kernel_bounds = self.kernel.get_bounds()
        bounds = []
        for i in range(len(kernel_bounds[0])):
            bounds.append(tuple([kernel_bounds[0][i], kernel_bounds[1][i]]))
        solutions = []
        solution_values = []
        for i in range(len(self.kernel.x0)):
            solution = self.algorithm(self.kernel.__costs__[i], self.kernel.x0[i], bounds=bounds)
            solution_values.append(solution.fun)
            solutions.append(solution)
        best_solution = solutions[np.argmin(solution_values)]
        x0 = best_solution.x
        solution = self.algorithm(self.kernel.__cost__, x0, bounds=bounds)
        self.solution = solution
        self.solutions = solutions
        return solution

    def get_constants(self):
        if self.solution is None:
            print('Solver not run.')
            return None
        spherical_solution = np.ones((self.kernel.n, self.kernel.d))
        spherical_solution[:, 1:] = self.solution.x.reshape(self.kernel.n, self.kernel.d - 1)
        cartesian_solution = sph2cart(spherical_solution)
        g = cartesian_solution.flatten()
        if self.kernel.lam == 0:
            s = np.zeros(self.kernel.n)
        else:
            s = -self.kernel.lam * np.linalg.inv(
                np.eye(self.kernel.j00_.shape[0]) + self.kernel.lam * self.kernel.j00_) @ \
                self.kernel.j01_ @ g
        left_side = np.zeros(self.kernel.a_inv.shape[0])
        left_side[:self.kernel.n] = s
        left_side[self.kernel.n:self.kernel.n + g.shape[0]] = g
        constants = self.kernel.a_inv @ left_side
        return constants

    def get_normals(self):
        if self.solution is None:
            print('Solver not run.')
            return None
        spherical_solution = np.ones((self.kernel.n, self.kernel.d))
        spherical_solution[:, 1:] = self.solution.x.reshape(self.kernel.n, self.kernel.d - 1)
        cartesian_solution = sph2cart(spherical_solution)
        normals = cartesian_solution
        return normals



# cython: language_level=3
# distutils: language=c++
import numpy as np
cimport numpy as cnp
import cython
from libc.math cimport acos, pi
from libc.math cimport sqrt
from libcpp.vector cimport vector

cnp.import_array()

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef dot(double[:] vector1, double[:] vector2):
    """
    Returns the dot product of two vectors.
    """
    cdef int i
    cdef double result = 0
    for i in range(vector1.shape[0]):
        result += vector1[i] * vector2[i]
    return result

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
def get_angles(double[:,:] vector1, double[:,:] vector2):
    """
    Returns the angle in degrees between the
    first and second vector.
    """
    cdef Py_ssize_t i = vector1.shape[0]
    cdef cnp.ndarray[cnp.float64_t, ndim=1] angles = np.zeros(i, dtype=np.float64)
    cdef angles_mv = angles
    cdef dotval, norm_u, norm_v, x, denom, ratio

    for ii in range(i):
        # dot(u,v)
        dotval = dot(vector1[ii,:], vector2[ii,:])

        # norm(u)
        norm_u = 0
        norm_v = 0
        # compute norms manually
        for x in vector1[ii,:]:
            norm_u += x*x
        for x in vector2[ii,:]:
            norm_v += x*x
        norm_u = sqrt(norm_u)
        norm_v = sqrt(norm_v)

        # protect against zero norm or out-of-bounds
        denom = norm_u * norm_v
        if denom < 1e-15:  # or some small epsilon to avoid division by zero
            angles_mv[ii] = 0.0  # or handle it gracefully
        else:
            # clamp the ratio to [-1, 1] to avoid numerical issues
            ratio = dotval / denom
            if ratio > 1.0:
                ratio = 1.0
            elif ratio < -1.0:
                ratio = -1.0

            angles_mv[ii] = acos(ratio) * (180.0 / pi)

    return angles


import numpy
import pyvista
import pymeshfix
from scipy.interpolate import splprep, splev
from svv.domain.routines.boolean import boolean
from svv.simulation.utils.extract_faces import extract_faces
from svv.utils.remeshing.remesh import remesh_surface_2d, remesh_surface


def smooth_junctions(mesh):
    """
    Smooth the junctions of a PolyData mesh using PyVista.
    Parameters:
    - mesh (pv.PolyData): The input surface mesh.
    Returns:
    - pv.PolyData: The mesh with the junctions smoothed.
    """
    # Extract the faces of the mesh
    faces, walls, caps, shared_boundaries = extract_faces(mesh, None)
    hsize = mesh.hsize
    # If there is only one wall face then proceed with smoothing
    if len(walls) == 1:
        # Smooth the mesh
        smoothed_mesh = walls[0].smooth_taubin(boundary_smoothing=True, normalize_coordinates=True)
        boundaries = smoothed_mesh.extract_feature_edges(non_manifold_edges=False, feature_edges=False,
                                                         manifold_edges=False, boundary_edges=True)
        boundaries = boundaries.split_bodies()
        caps = []
        for boundary in boundaries:
            cap = remesh_surface_2d(boundary, nosurf=True, hsiz=hsize)
            caps.append(cap)
        caps.insert(0, smoothed_mesh)
        model = pyvista.merge(caps)
        model.hsize = hsize
        return model, smoothed_mesh, caps
    else:
        return None, None, None



# Copyright (c) Stanford University, The Regents of the University of
#               California, and others.
#
# All Rights Reserved.
#
# See Copyright-SimVascular.txt for additional details.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

class OutflowBoundaryConditionType(object):
    RCR = "rcr"
    RESISTANCE = "resistance"
    CORONARY = "coronary"
    SV_TO_ONED = {'rcrt.dat': RCR, 'resistance.dat': RESISTANCE, 'cort.dat': CORONARY}
    ONED_TO_SV = {RCR: 'rcrt.dat', RESISTANCE: 'resistance.dat', CORONARY: 'cort.dat'}


class MaterialModel(object):
    LINEAR = "LINEAR"
    OLUFSEN = "OLUFSEN"


class Parameters():
    """ The Parameter class stores the input parameters for a 1D mesh generation.
    """

    class Units(object):
        MM = "mm"
        CM = "cm"

    def __init__(self):
        self.boundary_surfaces_dir = None
        self.output_directory = None
        self.centerlines_input_file = None
        self.centerlines_output_file = None
        self.compute_centerlines = True
        self.compute_mesh = False
        self.mesh_output_file = '1d_model.vtp'
        self.model_name = None
        self.model_order = None
        self.solver_output_file = None
        self.surface_model = None
        self.uniform_material = True
        self.wall_properties_input_file = None
        self.wall_properties_output_file = None

        self.outputformat = "TEXT"

        self.inflow_input_file = None
        self.inlet_face_input_file = None
        self.inlet_face_seg_num = 0

        self.uniform_bc = True
        self.outflow_bc_type = None
        self.outflow_bc_file = None

        self.outlet_face_names_file = None
        self.CENTERLINES_OUTLET_FILE_NAME = "centerlines_outlets.dat"

        # Mesh size in a sub-segment.
        self.element_size = 0.01

        # Min number of elements for a sub-segment.
        self.min_num_elems = 5

        # Min number of sub-segments in a vessel master-segment.
        self.seg_min_num = 1

        # Sub-segment size in a vessel-master segment.
        self.seg_size = 9999

        # Adaptive sub-segment size according to centerline area change?
        self.seg_size_adaptive = False

        self.time_step = 0.001
        self.num_time_steps = 1000
        self.save_data_freq = 1

        # Units conversion from mm to cgs.
        self.units = self.Units.CM
        self.lcoef = 1.0
        self.Acoef = 1.0

        # Physical parameters.
        self.density = 1.055
        self.material_model = MaterialModel.OLUFSEN
        self.viscosity = 0.04
        self.olufsen_material_k1 = 0.0
        self.olufsen_material_k2 = -22.5267
        self.olufsen_material_k3 = 1.0e7
        self.olufsen_material_exponent = 1.0
        self.olufsen_material_pressure = 0.0
        self.linear_material_ehr = 1e7
        self.linear_material_pressure = 0.0

        # Properties for the 0D model
        self.uniform_0d_element_type = True  # True if all 0D elements have the same 0d element type
        self.zerod_element_type = 'R'

    def set_units(self, units):
        units = units.lower()
        units_ok = True

        if units == self.Units.MM:
            self.lcoef = 0.1
            self.Acoef = 0.01
        elif units == self.Units.CM:
            self.lcoef = 1.0
            self.Acoef = 1.0
        else:
            units_ok = False

        return units_ok
